from django.test import TestCase
from querybuilder.helpers import value_for_keypath, set_value_for_keypath


class HelperTest(TestCase):
    """
    Tests the helper functions
    """
    def test_value_for_keypath(self):
        """
        Tests all cases of value_for_keypath
        """
        self.assertEqual({}, value_for_keypath({}, ''))
        self.assertIsNone(value_for_keypath({}, 'fake'))
        self.assertIsNone(value_for_keypath({}, 'fake.path'))
        self.assertEqual({'fruit': 'apple'}, value_for_keypath({'fruit': 'apple'}, ''))
        self.assertEqual('apple', value_for_keypath({'fruit': 'apple'}, 'fruit'))
        self.assertIsNone(value_for_keypath({'fruit': 'apple'}, 'fake'))
        self.assertIsNone(value_for_keypath({'fruit': 'apple'}, 'fake.path'))
        self.assertEqual(
            {'fruits': {'apple': 'red', 'banana': 'yellow'}},
            value_for_keypath({'fruits': {'apple': 'red', 'banana': 'yellow'}}, '')
        )
        self.assertEqual(
            {'apple': 'red', 'banana': 'yellow'},
            value_for_keypath({'fruits': {'apple': 'red', 'banana': 'yellow'}}, 'fruits')
        )
        self.assertEqual('red', value_for_keypath({'fruits': {'apple': 'red', 'banana': 'yellow'}}, 'fruits.apple'))
        self.assertEqual(
            {'color': 'red', 'taste': 'good'},
            value_for_keypath({'fruits': {'apple': {'color': 'red', 'taste': 'good'}}}, 'fruits.apple')
        )
        self.assertEqual(
            'red',
            value_for_keypath({'fruits': {'apple': {'color': 'red', 'taste': 'good'}}}, 'fruits.apple.color')
        )
        self.assertEqual(
            'good',
            value_for_keypath({'fruits': {'apple': {'color': 'red', 'taste': 'good'}}}, 'fruits.apple.taste')
        )

    def test_set_value_for_keypath(self):
        """
        Tests all cases of set_value_for_keypath
        """
        self.assertIsNone(set_value_for_keypath({}, '', None))
        self.assertIsNone(set_value_for_keypath({}, '', 'test value'))
        self.assertIsNone(set_value_for_keypath({'fruit': 'apple'}, '', None))
        self.assertIsNone(set_value_for_keypath({'fruit': 'apple'}, '', 'test value'))
        self.assertEqual({'fruit': None}, set_value_for_keypath({'fruit': 'apple'}, 'fruit', None))
        self.assertEqual({'fruit': 'test value'}, set_value_for_keypath({'fruit': 'apple'}, 'fruit', 'test value'))
        self.assertIsNone(set_value_for_keypath({'fruit': 'apple'}, 'fake', None))
        self.assertIsNone(set_value_for_keypath({'fruit': 'apple'}, 'fake', 'test value'))
        self.assertIsNone(set_value_for_keypath({'fruit': 'apple'}, 'fake.fake', 'test value'))
        self.assertEqual(
            {'fruit': {'apple': 'green'}},
            set_value_for_keypath({'fruit': {'apple': 'red'}}, 'fruit.apple', 'green')
        )
        self.assertEqual(
            {'fruit': {'apple': None}},
            set_value_for_keypath({'fruit': {'apple': 'red'}}, 'fruit.apple', None)
        )
        self.assertIsNone(set_value_for_keypath({'fruit': {'apple': {'color': 'red'}}}, 'fruit.apple.fake', 'green'))
        self.assertEqual(
            {'fruit': {'apple': {'color': 'green'}}},
            set_value_for_keypath({'fruit': {'apple': {'color': 'red'}}}, 'fruit.apple.color', 'green')
        )
        self.assertEqual(
            {'fruit': {'apple': {'color': {'puppies': {'count': 10, 'breed': 'boxers'}}}}},
            set_value_for_keypath(
                {'fruit': {'apple': {'color': 'red'}}},
                'fruit.apple.color',
                {'puppies': {'count': 10, 'breed': 'boxers'}}
            )
        )
        self.assertEqual(
            {'fruit': {'apple': {'color': 'red', 'animals': {'puppies': {'count': 10, 'breed': 'boxers'}}}}},
            set_value_for_keypath(
                {'fruit': {'apple': {'color': 'red'}}},
                'fruit.apple.animals',
                {'puppies': {'count': 10, 'breed': 'boxers'}},
                create_if_needed=True
            )
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Account',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('first_name', models.CharField(max_length=64)),
                ('last_name', models.CharField(max_length=64)),
            ],
        ),
        migrations.CreateModel(
            name='MetricRecord',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('other_value', models.FloatField(default=0)),
                ('data', jsonfield.fields.JSONField()),
            ],
        ),
        migrations.CreateModel(
            name='Order',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('revenue', models.FloatField(null=True)),
                ('margin', models.FloatField()),
                ('margin_percent', models.FloatField()),
                ('time', models.DateTimeField()),
                ('account', models.ForeignKey(to='tests.Account')),
            ],
        ),
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('email', models.CharField(max_length=256)),
            ],
        ),
        migrations.AddField(
            model_name='account',
            name='user',
            field=models.OneToOneField(to='tests.User'),
        ),
    ]



from querybuilder.fields import CountField, AvgField, MaxField, MinField, StdDevField, SumField, VarianceField
from querybuilder.query import Query
from querybuilder.tests.models import Order, User
from querybuilder.tests.query_tests import QueryTestCase, get_comparison_str


class AggregateTest(QueryTestCase):
    def test_count_id(self):
        query = Query().from_table(
            table='test_table',
            fields=[
                CountField('id')
            ]
        )
        query_str = query.get_sql()
        expected_query = 'SELECT COUNT(test_table.id) AS id_count FROM test_table'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_count_all(self):
        query = Query().from_table(
            table='test_table',
            fields=[
                CountField('*')
            ]
        )
        query_str = query.get_sql()
        expected_query = 'SELECT COUNT(test_table.*) AS all_count FROM test_table'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_count_id_alias(self):
        query = Query().from_table(
            table='test_table',
            fields=[{
                'num': CountField('id')
            }]
        )
        query_str = query.get_sql()
        expected_query = 'SELECT COUNT(test_table.id) AS num FROM test_table'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_avg(self):
        query = Query().from_table(
            table=Order,
            fields=[
                AvgField('margin')
            ]
        )
        query_str = query.get_sql()
        expected_query = 'SELECT AVG(tests_order.margin) AS margin_avg FROM tests_order'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_max_field(self):
        """
        Verifies that the MAX function is generated correctly in a query
        """
        query = Query().from_table(
            table=Order,
            fields=[
                MaxField('margin')
            ]
        )
        query_str = query.get_sql()
        expected_query = 'SELECT MAX(tests_order.margin) AS margin_max FROM tests_order'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_min_field(self):
        """
        Verifies that the MinField generates correct MIN sql
        """
        query = Query().from_table(
            table=Order,
            fields=[
                MinField('margin')
            ]
        )
        query_str = query.get_sql()
        expected_query = 'SELECT MIN(tests_order.margin) AS margin_min FROM tests_order'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_stddev(self):
        query = Query().from_table(
            table=Order,
            fields=[
                StdDevField('margin'),
            ]
        )
        query_str = query.get_sql()
        expected_query = 'SELECT STDDEV(tests_order.margin) AS margin_stddev FROM tests_order'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_sum_field(self):
        """
        Tests that the SumField generates correct sql
        """
        query = Query().from_table(
            table=Order,
            fields=[
                SumField('margin')
            ]
        )
        query_str = query.get_sql()
        expected_query = 'SELECT SUM(tests_order.margin) AS margin_sum FROM tests_order'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_variance(self):
        query = Query().from_table(
            table=Order,
            fields=[
                VarianceField('margin')
            ]
        )
        query_str = query.get_sql()
        expected_query = 'SELECT VARIANCE(tests_order.margin) AS margin_variance FROM tests_order'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_count(self):
        query = Query().from_table(
            User,
            ['one', 'two']
        ).order_by('id')
        received = query.count()
        expected = len(User.objects.all())
        self.assertEqual(
            received,
            expected,
            'Expected {0} but received {1}'.format(
                expected,
                received
            )
        )
        self.assertEqual(query.get_count_query().get_sql(), 'SELECT COUNT(tests_user.*) AS all_count FROM tests_user')

        # Make sure the copy didn't modify the original
        self.assertEqual(len(query.tables[0].fields), 2)
        self.assertEqual(len(query.sorters), 1)

    def test_max(self):
        """
        Tests that the max() function properly gets the max value from the query
        """
        query = Query().from_table(
            User
        )
        received = query.max('id')
        expected = User.objects.all().order_by('-id')[0].id
        self.assertEqual(
            received,
            expected,
            'Expected {0} but received {1}'.format(
                expected,
                received
            )
        )

    def test_min(self):
        """
        Tests the min() function
        """
        query = Query().from_table(
            User
        )
        received = query.min('id')
        expected = User.objects.all().order_by('id')[0].id
        self.assertEqual(
            received,
            expected,
            'Expected {0} but received {1}'.format(
                expected,
                received
            )
        )

    def test_sum(self):
        """
        Tests the sum() function on a field
        """
        query = Query().from_table(
            Order
        )
        received = query.sum('margin')
        expected = sum([order.margin for order in Order.objects.all()])
        self.assertEqual(
            received,
            expected,
            'Expected {0} but received {1}'.format(
                expected,
                received
            )
        )

    def test_average(self):
        """
        Tests the avg function
        """
        query = Query().from_table(
            Order
        )
        received = query.avg('margin')
        items = [order.margin for order in Order.objects.all()]
        average = sum(items) / len(items)
        expected = average
        self.assertEqual(
            received,
            expected,
            'Expected {0} but received {1}'.format(
                expected,
                received
            )
        )



from copy import deepcopy

from django.db import connection as default_django_connection
from django.db.models import Q
from django.db.models.query import QuerySet
from django.db.models.constants import LOOKUP_SEP
import six

from querybuilder.fields import FieldFactory, CountField, MaxField, MinField, SumField, AvgField
from querybuilder.helpers import set_value_for_keypath
from querybuilder.tables import TableFactory, ModelTable, QueryTable

try:
    from django.apps import apps
    get_model = apps.get_model
except ImportError:
    from django.db.models import get_model


class Join(object):
    """
    Represents the JOIN clauses of a Query. The join can be of any join type.
    """

    def __init__(self, right_table=None, fields=None, condition=None, join_type='JOIN',
                 schema=None, left_table=None, owner=None, extract_fields=True,
                 prefix_fields=True, field_prefix=None):
        """
        Initializes the default values and assigns any passed params

        :param right_table: The table being joined with. This can be a string of the table
            name, a dict of {'alias': table}, or a ``Table`` instance
        :type right_table: str or dict or :class:`Table <querybuilder.tables.Table>`

        :param fields: The fields to select from ``table``. Defaults to None. This can be
            a single field, a tuple of fields, or a list of fields. Each field can be a string
            or ``Field`` instance
        :type fields: str or tuple or list or :class:`Field <querybuilder.fields.Field>`

        :param condition: The join condition specifying the fields being joined. If the two tables being
            joined are instances of ``ModelTable`` then the condition should be created automatically.
        :type condition: str

        :param join_type: The type of join (JOIN, LEFT JOIN, INNER JOIN, etc). Defaults to 'JOIN'
        :type join_type: str

        :param schema: This is not implemented, but it will be a string of the db schema name
        :type schema: str

        :param left_table: The left table being joined with. This can be a string of the table
            name, a dict of {'alias': table}, or a ``Table`` instance. Defaults to the first table
            in the query.
        :type left_table: str or dict or :class:`Table <querybuilder.tables.Table>` or None

        :param owner: A reference to the query managing this Join object
        :type owner: :class:`Query <querybuilder.query.Query>`

        :param extract_fields: If True and joining with a ``ModelTable``, then '*'
            fields will be converted to individual fields for each column in the table. Defaults
            to True.
        :type extract_fields: bool

        :param prefix_fields: If True, then the joined table will have each of its field names
            prefixed with the field_prefix. If no field_prefix is specified, a name will be
            generated based on the join field name. This is usually used with nesting results
            in order to create models in python or javascript. Defaults to True.
        :type prefix_fields: bool

        :param field_prefix: The field prefix to be used in front of each field name if prefix_fields
            is set to True. If no field_prefix is set, one will be automatically created based on
            the join field name.
        :type field_prefix: str
        """
        self.owner = owner
        self.left_table = None
        self.right_table = None
        # self.table_join_name = None
        self.prefix_fields = prefix_fields
        self.condition = condition
        self.join_type = join_type
        self.schema = schema

        self.set_left_table(left_table=left_table)
        self.set_right_table(TableFactory(
            table=right_table,
            fields=fields,
            extract_fields=extract_fields,
            prefix_fields=prefix_fields,
            owner=self.owner,
            field_prefix=field_prefix,
        ))

    def get_sql(self):
        """
        Generates the JOIN sql for the join tables and join condition

        :rtype: str
        :return: the JOIN sql for the join tables and join condition
        """
        return '{0} {1} ON {2}'.format(self.join_type, self.right_table.get_sql(), self.get_condition())

    def set_left_table(self, left_table=None):
        """
        Sets the left table for this join clause. If no table is specified, the first table
        in the query will be used

        :type left_table: str or dict or :class:`Table <querybuilder.tables.Table>` or None
        :param left_table: The left table being joined with. This can be a string of the table
            name, a dict of {'alias': table}, or a ``Table`` instance. Defaults to the first table
            in the query.
        """
        if left_table:
            self.left_table = TableFactory(
                table=left_table,
                owner=self.owner,
            )
        else:
            self.left_table = self.get_left_table()

    def get_left_table(self):
        """
        Returns the left table if one was specified, otherwise the first
        table in the query is returned

        :rtype: :class:`Table <querybuilder.tables.Table>`
        :return: the left table if one was specified, otherwise the first table in the query
        """
        if self.left_table:
            return self.left_table
        if len(self.owner.tables):
            return self.owner.tables[0]

    def set_right_table(self, table):
        """
        Sets the right table for this join clause and try to automatically set the condition
        if one isn't specified
        """
        self.right_table = table
        if self.left_table is None:
            return

        # find table prefix
        if type(self.left_table) is ModelTable and type(self.right_table) is ModelTable:
            # loop through fields to find the field for this model

            # check if this join type is for a related field
            for field in self.left_table.model._meta.get_all_related_objects():
                related_model = field.model
                if hasattr(field, 'related_model'):
                    related_model = field.related_model
                if related_model == self.right_table.model:
                    if self.right_table.field_prefix is None:
                        self.right_table.field_prefix = field.get_accessor_name()
                        if len(self.right_table.field_prefix) > 4 and self.right_table.field_prefix[-4:] == '_set':
                            self.right_table.field_prefix = self.right_table.field_prefix[:-4]
                    return

            # check if this join type is for a foreign key
            for field in self.left_table.model._meta.fields:
                if (
                    field.get_internal_type() == 'OneToOneField' or
                    field.get_internal_type() == 'ForeignKey'
                ):
                    if field.rel.to == self.right_table.model:
                        if self.right_table.field_prefix is None:
                            self.right_table.field_prefix = field.name
                        return

    def get_condition(self):
        """
        Determines the condition to be used in the condition part of the join sql.

        :return: The condition for the join clause
        :rtype: str or None
        """
        if self.condition:
            return self.condition

        if type(self.right_table) is ModelTable and type(self.right_table) is ModelTable:
            # loop through fields to find the field for this model

            # check if this join type is for a related field
            for field in self.right_table.model._meta.get_all_related_objects():
                related_model = field.model
                if hasattr(field, 'related_model'):
                    related_model = field.related_model
                if related_model == self.left_table.model:
                    table_join_field = field.field.column
                    # self.table_join_name = field.get_accessor_name()
                    condition = '{0}.{1} = {2}.{3}'.format(
                        self.right_table.get_identifier(),
                        self.right_table.model._meta.pk.name,
                        self.left_table.get_identifier(),
                        table_join_field,
                    )
                    return condition

            # check if this join type is for a foreign key
            for field in self.right_table.model._meta.fields:
                if (
                    field.get_internal_type() == 'OneToOneField' or
                    field.get_internal_type() == 'ForeignKey'
                ):
                    if field.rel.to == self.left_table.model:
                        table_join_field = field.column
                        # self.table_join_name = field.name
                        condition = '{0}.{1} = {2}.{3}'.format(
                            self.right_table.get_identifier(),
                            table_join_field,
                            self.left_table.get_identifier(),
                            self.left_table.model._meta.pk.name
                        )
                        return condition
        return None


# TODO: make an expression table and expression field maybe
class Expression(object):

    def __init__(self, str):
        self.str = str


class Where(object):
    """
    Represents the WHERE clause of a Query. The filter data is contained inside of django
    Q objects and methods are provided to interface with them.

    Properties:

        arg_index: int
            The numeric index that is automatically assigned to query parameters

        arg_prefix: str
            A prefix for the arg names used to namespace inner queries. This is set
            by the Query object

        args: dict
            A dictionary mapping the arg keys to the actual values. This is the data
            that is passed into cursor.execute

        wheres: Q
            A django Q object that can contain many nested Q objects that are used to
            determine all of the where conditions and nested where conditions
    """

    comparison_map = {
        'exact': '=',
        'eq': '=',
        'gt': '>',
        'gte': '>=',
        'lt': '<',
        'lte': '<=',
        'contains': 'LIKE',
        'startswith': 'LIKE',
        'in': 'IN',
    }

    def __init__(self):
        """
        Initializes default variables
        """
        self.arg_index = 0
        self.arg_prefix = ''
        self.args = {}
        self.wheres = Q()

    def get_sql(self):
        """
        Builds and returns the WHERE portion of the sql

        :return: the WHERE portion of the sql
        :rtype: str
        """
        # reset arg index and args
        self.arg_index = 0
        self.args = {}

        # build the WHERE sql portion if needed
        if len(self.wheres):
            where = self.build_where_part(self.wheres)
            return 'WHERE {0} '.format(where)
        return ''

    def get_condition_operator(self, operator):
        """
        Gets the comparison operator from the Where class's comparison_map

        :return: the comparison operator from the Where class's comparison_map
        :rtype: str
        """
        return Where.comparison_map.get(operator, None)

    def get_condition_value(self, operator, value):
        """
        Gets the condition value based on the operator and value

        :param operator: the condition operator name
        :type operator: str
        :param value: the value to be formatted based on the condition operator
        :type value: object

        :return: the comparison operator from the Where class's comparison_map
        :rtype: str
        """
        if operator == 'contains':
            value = '%{0}%'.format(value)
        elif operator == 'startswith':
            value = '{0}%'.format(value)
        return value

    def build_where_part(self, wheres):
        """
        Recursive method that builds the where parts. Any Q objects that have children will
        also be built with ``self.build_where_part()``

        :rtype: str
        :return: The composed where string
        """
        where_parts = []

        # loop through each child of the Q condition
        for where in wheres.children:

            # if this child is another Q object, recursively build the where part
            if type(where) is Q:
                where_parts.append(self.build_where_part(where))
            elif type(where) is tuple:
                # build the condition for this where part
                # get the field name and value
                field_name = where[0]
                value = where[1]

                # set the default operator
                operator_str = 'eq'
                operator = '='

                # break apart the field name on double underscores
                # TODO: do not convert the first double underscore to a .
                field_parts = field_name.split('__')
                if len(field_parts) > 1:
                    # get the operator based on the last element split from the double underscores
                    operator_str = field_parts[-1]
                    operator = self.get_condition_operator(operator_str)
                    if operator is None:
                        operator = '='
                        field_name = '.'.join(field_parts)
                    else:
                        field_name = '.'.join(field_parts[:-1])

                    # if there is more than one double underscore, make the first one a dot
                    trimmed_field_parts = field_name.split('.')
                    if len(trimmed_field_parts) > 2:
                        field_name = '{0}.{1}'.format(trimmed_field_parts[0], '__'.join(trimmed_field_parts[1:]))

                # check if we are comparing to null
                if value is None:
                    # change the operator syntax to IS
                    operator = 'IS'

                # set up the condition string format
                condition = '{0} {1} ?'.format(field_name, operator)

                # apply the NOT if this condition is negated
                if wheres.negated:
                    condition = 'NOT({0})'.format(condition)

                # check if this value is multiple values
                if operator_str == 'in':

                    # make sure value is a list
                    if type(value) is not list:

                        # convert to string in case it is a number
                        value = str(value)

                        # split on commas
                        value = value.split(',')

                    # Ensure that we have a value in the list
                    if len(value) == 0:
                        value = [None]

                    if type(value) is Expression:
                        condition = condition.replace('?', value.str)
                    else:
                        # assign each query param to a named arg
                        named_args = []
                        for value_item in value:
                            named_arg = self.set_arg(value_item)
                            named_args.append('%({0})s'.format(named_arg))
                        # replace the ? in the query with the arg placeholder
                        condition = condition.replace('?', '({0})'.format(','.join(named_args)), 1)
                else:
                    # get the value based on the operator
                    value = self.get_condition_value(operator_str, value)

                    if type(value) is Expression:
                        condition = condition.replace('?', value.str)
                    else:
                        named_arg = self.set_arg(value)
                        # replace the ? in the query with the arg placeholder
                        condition = condition.replace('?', '%({0})s'.format(named_arg), 1)

                # add the condition to the where sql
                where_parts.append(condition)

        # join all where parts together
        joined_parts = ' {0} '.format(wheres.connector).join(where_parts)

        # wrap the where parts in parentheses
        return '({0})'.format(joined_parts)

    def set_arg(self, value):
        """
        Set the query param in self.args based on the prefix and arg index
        and auto increment the arg_index

        :return: the string placeholder for the arg
        :rtype: str
        """
        named_arg = '{0}A{1}'.format(self.arg_prefix, self.arg_index)
        self.args[named_arg] = value
        self.arg_index += 1
        return named_arg


class Group(object):
    """
    Represents a group by clause used in a Query
    """

    def __init__(self, field=None, table=None):
        """
        :param field: This can be a string of a field name, a dict of {'alias': field}, or
            a ``Field`` instance
        :type field: str or dict or :class:`Field <querybuilder.fields.Field>`

        :param table: Optional. This can be a string of a table name, a dict of {'alias': table}, or
            a ``Table`` instance. A table only needs to be supplied in more complex queries where
            the field name is ambiguous.
        :type table: str or dict or :class:`Table <querybuilder.tables.Table>`
        """
        self.field = FieldFactory(field)
        self.table = TableFactory(table)
        if self.table and self.field.table is None:
            self.field.set_table(self.table)

    def get_name(self):
        """
        Gets the name to reference the grouped field

        :return: the name to reference the grouped field
        :rtype: str
        """
        return self.field.get_identifier()


class Sorter(object):
    """
    Used internally by the Query class to set ORDER BY clauses on the query.
    """

    def __init__(self, field=None, table=None, desc=False):
        """
        Initializes the instance variables

        :type field: str or dict or :class:`Field <querybuilder.fields.Field>`
        :param field: This can be a string of a field name, a dict of {'alias': field}, or
            a ``Field`` instance

        :type table: str or dict or :class:`Table <querybuilder.tables.Table>`
        :param table: Optional. This can be a string of a table name, a dict of {'alias': table}, or
            a ``Table`` instance. A table only needs to be supplied in more complex queries where
            the field name is ambiguous.

        :param desc: Set to True to sort by this field in DESC order or False to sort by this field
            in ASC order. Defaults to False.
        :type desc: bool
        """
        self.desc = desc
        self.field = FieldFactory(field)
        self.table = TableFactory(table)

        # if the field is not associated with a table but a table was
        # passed in, set the field's table to the passed table
        if self.table and self.field.table is None:
            self.field.set_table(self.table)

        # if the specified field is a string with '-' at the beginning
        # the '-' needs to be removed and this sorter needs to be
        # set to desc
        if isinstance(self.field.field, six.string_types) and str(self.field.field[0]) == '-':
            self.desc = True
            self.field.field = self.field.field[1:]
            self.field.name = self.field.name[1:]

    def get_name(self, use_alias=True):
        """
        Gets the name to reference the sorted field

        :return: the name to reference the sorted field
        :rtype: str
        """
        if self.desc:
            direction = 'DESC'
        else:
            direction = 'ASC'

        if use_alias:
            return '{0} {1}'.format(self.field.get_identifier(), direction)
        return '{0} {1}'.format(self.field.get_select_sql(), direction)


class Limit(object):
    """
    Used internally by the Query class to set a limit and/or offset on the query.
    """

    def __init__(self, limit=None, offset=None):
        """
        Initializes the instance variables

        :param limit: the number of rows to return
        :type limit: int

        :param offset: the number of rows to start returning rows from
        :type limit: int
        """
        self.limit = limit
        self.offset = offset

    def get_sql(self):
        """
        Generates the sql used for the limit clause of a Query

        :return: the sql for the limit clause of a Query
        :rtype: str
        """
        sql = ''
        if self.limit and self.limit > 0:
            sql += 'LIMIT {0} '.format(self.limit)
        if self.offset and self.offset > 0:
            sql += 'OFFSET {0} '.format(self.offset)
        return sql


class Query(object):
    """
    A Query instance represents an actual query that will be executed. It provides
    methods for selecting fields from tables, inner queries, joins, filtering,
    limiting, and sorting. Different types of queries can be executed, such as:
    select, update, delete, create, and explain.

    Properties:

        sql: str
            The query generated by calling ``self.get_sql()`` This is used for
            caching purposes.

        tables: list of Table
            A list of ``Table`` instances this query is selecting from

        joins: list of Join
            A list of ``Join`` instances this query is joining on

        _where: Where
            A ``Where`` instance containing filtering data for this query

        groups: list of Group
            A list of ``Group`` instances that determine the GROUP BY clause for this query

        sorters: list of Sorter
            A list of ``Sorter`` instances that determine the ORDER BY clause for this query

        _limit: Limit
            An instance of ``Limit`` This will only exist if a limit has been specified for the query

        table_prefix: str
            A ``str`` that determines how to prefix inner queries of this query

    """
    enable_safe_limit = False
    safe_limit = 1000

    def init_defaults(self):
        """
        Sets the default values for this instance
        """
        self.sql = ''
        self.tables = []
        self.joins = []
        self._where = Where()
        self.groups = []
        self.sorters = []
        self._limit = None
        self.table_prefix = ''
        self.is_inner = False
        self.with_tables = []
        self._distinct = False
        self.field_names = []
        self.field_names_pk = None
        self.values = []

    def __init__(self, connection=None):
        """
        Initializes this instance by calling ``self.init_defaults``

        :type connection: :class:`DatabaseWrapper <django:django.db.backends.BaseDatabaseWrapper>`
        :parameter connection: A Django database connection. This can be used to connect to
            databases other than your default database.

        .. code-block:: python

            from django.db import connections
            from querybuilder.query import Query

            Query(connections.all()[0]).from_table('auth_user').count()
            # 15L
            Query(connections.all()[1]).from_table('auth_user').count()
            # 223L

        """
        self.connection = connection or default_django_connection

        self.init_defaults()

    def get_cursor(self):
        """
        Get a cursor for the Query's connection

        :rtype: :class:`CursorDebugWrapper <django:django.db.backends.util.CursorDebugWrapper>`
        :returns: A database cursor
        """
        return self.connection.cursor()

    def from_table(self, table=None, fields='*', schema=None, **kwargs):
        """
        Adds a ``Table`` and any optional fields to the list of tables
        this query is selecting from.

        :type table: str or dict or :class:`Table <querybuilder.tables.Table>`
            or :class:`Query <querybuilder.query.Query>` or
            :class:`ModelBase <django:django.db.models.base.ModelBase>`
        :param table: The table to select fields from. This can be a string of the table
            name, a dict of {'alias': table}, a ``Table`` instance, a Query instance, or a
            django Model instance

        :type fields: str or tuple or list or Field
        :param fields: The fields to select from ``table``. Defaults to '*'. This can be
            a single field, a tuple of fields, or a list of fields. Each field can be a string
            or ``Field`` instance

        :type schema: str
        :param schema: This is not implemented, but it will be a string of the db schema name

        :param kwargs: Any additional parameters to be passed into the constructor of ``TableFactory``

        :return: self
        :rtype: :class:`Query <querybuilder.query.Query>`
        """
        # self.mark_dirty()

        self.tables.append(TableFactory(
            table=table,
            fields=fields,
            schema=schema,
            owner=self,
            **kwargs
        ))

        return self

    def insert_into(self, table=None, field_names=None, values=None, **kwargs):
        """
        Bulk inserts a list of values into a table

        :type table: str or dict or :class:`Table <querybuilder.tables.Table>`
            or :class:`Query <querybuilder.query.Query>` or :class:`ModelBase <django:django.db.models.base.ModelBase>`
        :param table: The table to select fields from. This can be a string of the table
            name, a dict of {'alias': table}, a ``Table`` instance, a Query instance, or a
            django Model instance

        :type field_names: list
        :param field_names: A list of ordered field names that relate to the data in the values list

        :type values: list of list
        :param values: A list each values list with the values in the same order as the field names

        :param kwargs: Any additional parameters to be passed into the constructor of ``TableFactory``

        :return: self
        :rtype: :class:`Query <querybuilder.query.Query>`
        """
        table = TableFactory(
            table=table,
            **kwargs
        )
        self.tables.append(table)

        self.field_names = field_names
        self.values = values

        return self

    def update_table(self, table=None, field_names=None, values=None, pk=None, **kwargs):
        """
        Bulk updates rows in a table

        :type table: str or dict or :class:`Table <querybuilder.tables.Table>`
            or :class:`Query <querybuilder.query.Query>` or :class:`ModelBase <django:django.db.models.base.ModelBase>`
        :param table: The table to select fields from. This can be a string of the table
            name, a dict of {'alias': table}, a ``Table`` instance, a Query instance, or a
            django Model instance

        :type field_names: list
        :param field_names: A list of ordered field names that relate to the data in the values list

        :type values: list of list
        :param values: A list each values list with the values in the same order as the field names

        :type pk: int
        :param pk: The name of the primary key in the table and field_names

        :param kwargs: Any additional parameters to be passed into the constructor of ``TableFactory``

        :rtype: :class:`Query <querybuilder.query.Query>`
        :return: self
        """
        table = TableFactory(
            table=table,
            **kwargs
        )
        self.tables.append(table)

        self.field_names = field_names
        self.values = values
        self.field_names_pk = pk

    # TODO: add docs
    # TODO: add tests for custom with clauses
    def with_query(self, query=None, alias=None):
        """

        :return: self
        :rtype: :class:`Query <querybuilder.query.Query>`
        """
        self.with_tables.append(TableFactory(query, alias=alias))
        return self

    def join(self, right_table=None, fields=None, condition=None, join_type='JOIN',
             schema=None, left_table=None, extract_fields=True, prefix_fields=False, field_prefix=None,
             allow_duplicates=False):
        """
        Joins a table to another table based on a condition and adds fields from the joined table
        to the returned fields.

        :type right_table: str or dict or :class:`Table <querybuilder.tables.Table>`
        :param right_table: The table being joined with. This can be a string of the table
            name, a dict of {'alias': table}, or a ``Table`` instance

        :type fields: str or tuple or list or :class:`Field <querybuilder.fields.Field>`
        :param fields: The fields to select from ``right_table``. Defaults to `None`. This can be
            a single field, a tuple of fields, or a list of fields. Each field can be a string
            or ``Field`` instance

        :type condition: str
        :param condition: The join condition specifying the fields being joined. If the two tables being
            joined are instances of ``ModelTable`` then the condition should be created automatically.

        :type join_type: str
        :param join_type: The type of join (JOIN, LEFT JOIN, INNER JOIN, etc). Defaults to 'JOIN'

        :type schema: str
        :param schema: This is not implemented, but it will be a string of the db schema name

        :type left_table: str or dict or Table
        :param left_table: The left table being joined with. This can be a string of the table
            name, a dict of {'alias': table}, or a ``Table`` instance. Defaults to the first table
            in the query.

        :type extract_fields: bool
        :param extract_fields: If True and joining with a ``ModelTable``, then '*'
            fields will be converted to individual fields for each column in the table. Defaults
            to True.

        :type prefix_fields: bool
        :param prefix_fields: If True, then the joined table will have each of its field names
            prefixed with the field_prefix. If not field_prefix is specified, a name will be
            generated based on the join field name. This is usually used with nesting results
            in order to create models in python or javascript. Defaults to True.

        :type field_prefix: str
        :param field_prefix: The field prefix to be used in front of each field name if prefix_fields
            is set to True. If no field_prefix is set, one will be automatically created based on
            the join field name.

        :rtype: :class:`Query <querybuilder.query.Query>`
        :return: self
        """
        # self.mark_dirty()
        # TODO: fix bug when joining from simple table to model table with no condition
        # it assumes left_table.model

        # if there is no left table, assume the query's first table
        # TODO: add test for auto left table to replace old auto left table
        # if left_table is None and len(self.tables):
        #     left_table = self.tables[0]

        # left_table = TableFactory(left_table)
        # right_table = TableFactory(right_table)

        # create the join item
        new_join_item = Join(
            left_table=left_table,
            right_table=right_table,
            fields=fields,
            condition=condition,
            join_type=join_type,
            schema=schema,
            owner=self,
            extract_fields=extract_fields,
            prefix_fields=prefix_fields,
            field_prefix=field_prefix,
        )

        # check if this table is already joined upon
        # TODO: add test for this
        if allow_duplicates is False:
            for join_item in self.joins:
                if join_item.right_table.get_identifier() == new_join_item.right_table.get_identifier():
                    if join_item.left_table.get_identifier() == new_join_item.left_table.get_identifier():
                        return self

        self.joins.append(new_join_item)

        return self

    def join_left(self, right_table=None, fields=None, condition=None, join_type='LEFT JOIN',
                  schema=None, left_table=None, extract_fields=True, prefix_fields=False,
                  field_prefix=None, allow_duplicates=False):
        """
        Wrapper for ``self.join`` with a default join of 'LEFT JOIN'

        :type right_table: str or dict or :class:`Table <querybuilder.tables.Table>`
        :param right_table: The table being joined with. This can be a string of the table
            name, a dict of {'alias': table}, or a ``Table`` instance

        :type fields: str or tuple or list or :class:`Field <querybuilder.fields.Field>`
        :param fields: The fields to select from ``right_table``. Defaults to `None`. This can be
            a single field, a tuple of fields, or a list of fields. Each field can be a string
            or ``Field`` instance

        :type condition: str
        :param condition: The join condition specifying the fields being joined. If the two tables being
            joined are instances of ``ModelTable`` then the condition should be created automatically.

        :type join_type: str
        :param join_type: The type of join (JOIN, LEFT JOIN, INNER JOIN, etc). Defaults to 'JOIN'

        :type schema: str
        :param schema: This is not implemented, but it will be a string of the db schema name

        :type left_table: str or dict or :class:`Table <querybuilder.tables.Table>`
        :param left_table: The left table being joined with. This can be a string of the table
            name, a dict of {'alias': table}, or a ``Table`` instance. Defaults to the first table
            in the query.

        :type extract_fields: bool
        :param extract_fields: If True and joining with a ``ModelTable``, then '*'
            fields will be converted to individual fields for each column in the table. Defaults
            to True.

        :type prefix_fields: bool
        :param prefix_fields: If True, then the joined table will have each of its field names
            prefixed with the field_prefix. If not field_prefix is specified, a name will be
            generated based on the join field name. This is usually used with nesting results
            in order to create models in python or javascript. Defaults to True.

        :type field_prefix: str
        :param field_prefix: The field prefix to be used in front of each field name if prefix_fields
            is set to True. If no field_prefix is set, one will be automatically created based on
            the join field name.

        :return: self
        :rtype: :class:`Query <querybuilder.query.Query>`
        """
        return self.join(
            right_table=right_table,
            fields=fields,
            condition=condition,
            join_type=join_type,
            schema=schema,
            left_table=left_table,
            extract_fields=extract_fields,
            prefix_fields=prefix_fields,
            field_prefix=field_prefix,
            allow_duplicates=allow_duplicates
        )

    def where(self, q=None, where_type='AND', **kwargs):
        """
        Adds a where condition as a Q object to the query's ``Where`` instance.

        :type q: :class:`Q <django:django.db.models.Q>`
        :param q: A django ``Q`` instance. This will be added to the query's ``Where`` object. If no
            Q object is passed, the kwargs will be examined for params to be added to Q objects

        :param where_type: str
        :param where_type: The connection type of the where condition ('AND', 'OR')

        :return: self
        :rtype: :class:`Query <querybuilder.query.Query>`
        """
        # self.mark_dirty()
        if q is not None:
            self._where.wheres.add(q, where_type)
        if len(kwargs):
            for key, value in kwargs.items():
                q = Q(**{
                    key: value
                })
                self._where.wheres.add(q, where_type)
        return self

    def group_by(self, field=None, table=None, allow_duplicates=False):
        """
        Adds a group by clause to the query by adding a ``Group`` instance to the query's
        groups list

        :type field: str or dict or :class:`Field <querybuilder.fields.Field>`
        :param field: This can be a string of a field name, a dict of {'alias': field}, or
            a ``Field`` instance

        :type table: str or dict or :class:`Table <querybuilder.table.Table>`
        :param table: Optional. This can be a string of a table name, a dict of {'alias': table}, or
            a ``Table`` instance. A table only needs to be supplied in more complex queries where
            the field name is ambiguous.

        :return: self
        :rtype: :class:`Query <querybuilder.query.Query>`
        """
        new_group_item = Group(
            field=field,
            table=table,
        )

        if allow_duplicates is False:
            for group_item in self.groups:
                if group_item.field.get_identifier() == new_group_item.field.get_identifier():
                    return self

        self.groups.append(new_group_item)

        return self

    def order_by(self, field=None, table=None, desc=False):
        """
        Adds an order by clause to the query by adding a ``Sorter`` instance to the query's
        sorters list

        :type field: str or dict or :class:`Field <querybuilder.fields.Field>`
        :param field: This can be a string of a field name, a dict of {'alias': field}, or
            a ``Field`` instance

        :type table: str or dict or :class:`Table <querybuilder.table.Table>`
        :param table: Optional. This can be a string of a table name, a dict of {'alias': table}, or
            a ``Table`` instance. A table only needs to be supplied in more complex queries where
            the field name is ambiguous.

        :type desc: bool
        :param desc: Set to True to sort by this field in DESC order or False to sort by this field
            in ASC order. Defaults to False.

        :rtype: :class:`Query <querybuilder.query.Query>`
        :return: self
        """
        self.sorters.append(Sorter(
            field=field,
            table=table,
            desc=desc
        ))
        return self

    def limit(self, limit=None, offset=None):
        """
        Sets a limit and/or offset to the query to limit the number of rows returned.

        :type limit: int
        :param limit: The number of rows to return

        :type offset: int
        :param offset: The offset from the start of the record set where rows should start being returned

        :rtype: :class:`Query <querybuilder.query.Query>`
        :return: self
        """
        self._limit = Limit(
            limit=limit,
            offset=offset
        )
        return self

    def distinct(self, use_distinct=True):
        """
        Adds a distinct clause to the query

        :type use_distinct: bool
        :param use_distinct: Whether or not to include the distinct clause

        :rtype: :class:`Query <querybuilder.query.Query>`
        :return: self
        """
        self._distinct = use_distinct
        return self

    def distinct_on(self):
        raise NotImplementedError

    def check_name_collisions(self):
        """
        Checks if there are any tables referenced by the same identifier and updated the
        auto_alias accordingly. This is called when generating the sql for a query
        and should only be called internally.
        """
        table_index = 0
        table_names = {}
        for table in self.tables + self.with_tables:
            table_prefix = 'T{0}'.format(table_index)
            auto_alias = '{0}{1}'.format(self.table_prefix, table_prefix)

            identifier = table.get_identifier()
            if identifier is None or identifier in table_names:
                table.auto_alias = auto_alias
            table_names[identifier] = True

            # prefix inner query args and update self args
            if type(table) is QueryTable:
                table.query.prefix_args(auto_alias)
                table.query.table_prefix = auto_alias

            table_index += 1

    def prefix_args(self, prefix):
        """
        Adds an argument prefix to the query's ``Where`` object. This should only
        be called internally.
        """
        self._where.arg_prefix = prefix

    def get_sql(self, debug=False, use_cache=True):
        """
        Generates the sql for this query and returns the sql as a string.

        :type debug: bool
        :param debug: If True, the sql will be returned in a format that is easier to read and debug.
            Defaults to False

        :type use_cache: bool
        :param use_cache: If True, the query will returned the cached sql if it exists rather
            then generating the sql again. If False, the sql will be generated again. Defaults to True.

        :rtype: str
        :return: The generated sql for this query
        """
        # TODO: enable caching
        # if self.sql and use_cache and not debug:
        #     return self.sql

        # auto alias any naming collisions
        self.check_name_collisions()

        # if debugging, return the debug formatted sql
        if debug:
            return self.format_sql()

        # build each part of the query
        sql = ''
        sql += self.build_withs()
        sql += self.build_select_fields()
        sql += self.build_from_table()
        sql += self.build_joins()
        sql += self.build_where()
        sql += self.build_groups()
        sql += self.build_order_by()
        sql += self.build_limit()

        # remove any whitespace from the beginning and end of the sql
        self.sql = sql.strip()

        return self.sql

    def get_insert_sql(self, rows):
        field_names_sql = '({0})'.format(', '.join(self.get_field_names()))
        row_values = []
        sql_args = []
        for row in rows:
            placeholders = []
            for value in row:
                sql_args.append(value)
                placeholders.append('%s')
            row_values.append('({0})'.format(', '.join(placeholders)))
        row_values_sql = ', '.join(row_values)

        self.sql = 'INSERT INTO {0} {1} VALUES {2}'.format(
            self.tables[0].get_identifier(),
            field_names_sql,
            row_values_sql
        )

        return self.sql, sql_args

    def get_update_sql(self, rows):
        """
        Returns SQL UPDATE for rows ``rows``

        .. code-block:: sql

            UPDATE table_name
            SET
                field1 = new_values.field1
                field2 = new_values.field2
            FROM (
                VALUES
                    (1, 'value1', 'value2'),
                    (2, 'value1', 'value2')
            ) AS new_values (id, field1, field2)
            WHERE table_name.id = new_values.id;

        """
        field_names = self.get_field_names()
        pk = field_names[0]
        update_field_names = field_names[1:]

        num_columns = len(rows[0])
        if num_columns < 2:
            raise Exception('At least 2 fields must be passed to get_update_sql')

        all_null_indices = [
            all(row[index] is None for row in rows)
            for index in range(1, num_columns)
        ]

        field_names_sql = '({0})'.format(', '.join(field_names))

        row_values = []
        sql_args = []

        for row in rows:
            placeholders = []
            for value in row:
                sql_args.append(value)
                placeholders.append('%s')
            row_values.append('({0})'.format(', '.join(placeholders)))
        row_values_sql = ', '.join(row_values)

        # build field list for SET portion
        set_field_list = [
            '{0} = NULL'.format(field_name)
            if all_null_indices[idx] else '{0} = new_values.{0}'.format(field_name)
            for idx, field_name in enumerate(update_field_names)
        ]
        set_field_list_sql = ', '.join(set_field_list)

        self.sql = 'UPDATE {0} SET {1} FROM (VALUES {2}) AS new_values {3} WHERE {0}.{4} = new_values.{4}'.format(
            self.tables[0].get_identifier(),
            set_field_list_sql,
            row_values_sql,
            field_names_sql,
            pk
        )

        return self.sql, sql_args

    def format_sql(self):
        """
        Builds the sql in a format that is easy for humans to read and debug

        :return: The formatted sql for this query
        :rtype: str
        """
        # TODO: finish adding the other parts of the sql generation
        sql = ''

        # build SELECT
        select_segment = self.build_select_fields()
        select_segment = select_segment.replace('SELECT ', '', 1)
        fields = [field.strip() for field in select_segment.split(',')]
        sql += 'SELECT\n\t{0}\n'.format(',\n\t'.join(fields))

        # build FROM
        from_segment = self.build_from_table()
        from_segment = from_segment.replace('FROM ', '', 1)
        tables = [table.strip() for table in from_segment.split(',')]
        sql += 'FROM\n\t{0}\n'.format(',\n\t'.join(tables))

        # build ORDER BY
        order_by_segment = self.build_order_by()
        if len(order_by_segment):
            order_by_segment = order_by_segment.replace('ORDER BY ', '', 1)
            sorters = [sorter.strip() for sorter in order_by_segment.split(',')]
            sql += 'ORDER BY\n\t{0}\n'.format(',\n\t'.join(sorters))

        # build LIMIT
        limit_segment = self.build_limit()
        if len(limit_segment):
            if 'LIMIT' in limit_segment:
                limit_segment = limit_segment.replace('LIMIT ', 'LIMIT\n\t', 1)
                if 'OFFSET' in limit_segment:
                    limit_segment = limit_segment.replace('OFFSET ', '\nOFFSET\n\t', 1)
            elif 'OFFSET' in limit_segment:
                limit_segment = limit_segment.replace('OFFSET ', 'OFFSET\n\t', 1)
            sql += limit_segment

        return sql

    def get_field_names(self):
        """
        Builds a list of the field names for all tables and joined tables by calling
        ``get_field_names()`` on each table

        :return: list of field names
        :rtype: list of str
        """
        field_names = []
        for table in self.tables:
            field_names.extend(table.get_field_names())
        for join_item in self.joins:
            field_names.extend(join_item.right_table.get_field_names())
        return field_names

    def get_field_identifiers(self):
        """
        Builds a list of the field identifiers for all tables and joined tables by calling
        ``get_field_identifiers()`` on each table

        :return: list of field identifiers
        :rtype: list of str
        """
        field_identifiers = []
        for table in self.tables:
            field_identifiers += table.get_field_identifiers()
        for join_item in self.joins:
            field_identifiers += join_item.right_table.get_field_identifiers()
        return field_identifiers

    def build_insert_into(self):
        pass

    def build_withs(self):
        if self.is_inner:
            return ''

        withs = []
        for inner_query in self.with_tables + self.get_inner_queries():
            withs.append(inner_query.get_with_sql())
        if len(withs):
            withs.reverse()
            return 'WITH {0} '.format(', '.join(withs))
        return ''

    def get_inner_queries(self, query=None):
        inner_queries = []
        if query is None:
            query = self
        for table in query.tables:
            if type(table) is QueryTable:
                inner_queries.append(table)
                inner_queries += self.get_inner_queries(table.query)

        return inner_queries

    def build_select_fields(self):
        """
        Generates the sql for the SELECT portion of the query

        :return: the SELECT portion of the query
        :rtype: str
        """
        field_sql = []

        # get the field sql for each table
        for table in self.tables:
            field_sql += table.get_field_sql()

        # get the field sql for each join table
        for join_item in self.joins:
            field_sql += join_item.right_table.get_field_sql()

        # combine all field sql separated by a comma
        sql = 'SELECT {0}{1} '.format(self.get_distinct_sql(), ', '.join(field_sql))
        return sql

    def get_distinct_sql(self):
        if self._distinct:
            return 'DISTINCT '
        return ''

    def build_from_table(self):
        """
        Generates the sql for the FROM portion of the query

        :return: the FROM portion of the query
        :rtype: str
        """
        table_parts = []

        # get the table sql for each table
        for table in self.tables:
            sql = table.get_sql()
            if len(sql):
                table_parts.append(sql)

        # combine all table sql separated by a comma
        sql = 'FROM {0} '.format(', '.join(table_parts))

        return sql

    def build_joins(self):
        """
        Generates the sql for the JOIN portion of the query

        :return: the JOIN portion of the query
        :rtype: str
        """
        join_parts = []

        # get the sql for each join object
        for join_item in self.joins:
            join_parts.append(join_item.get_sql())

        # if there are any joins, combine them
        if len(join_parts):
            combined_joins = ' '.join(join_parts)
            return '{0} '.format(combined_joins)
        return ''

    def build_where(self):
        """
        Generates the sql for the WHERE portion of the query

        :return: the WHERE portion of the query
        :rtype: str
        """
        return self._where.get_sql()

    def build_groups(self):
        """
        Generates the sql for the GROUP BY portion of the query

        :return: the GROUP BY portion of the query
        :rtype: str
        """
        # check if there are any groupings
        if len(self.groups):
            groups = []

            # get the group sql for each grouping
            for group in self.groups:
                groups.append(group.get_name())
            return 'GROUP BY {0} '.format(', '.join(groups))
        return ''

    def build_order_by(self, use_alias=True):
        """
        Generates the sql for the ORDER BY portion of the query

        :type use_alias: bool
        :param use_alias: If True, the alias for the field will be used in the order by.
            This is an option before query windows do not use the alias. Defaults to True.

        :return: the ORDER BY portion of the query
        :rtype: str
        """
        # check if there are any sorters
        if len(self.sorters):
            sorters = []

            # get the sql for each sorter
            for sorter in self.sorters:
                sorters.append(sorter.get_name(use_alias=use_alias))
            return 'ORDER BY {0} '.format(', '.join(sorters))
        return ''

    def build_limit(self):
        """
        Generates the sql for the LIMIT and OFFSET portions of the query

        :return: the LIMIT and/or OFFSET portions of the query
        :rtype: str
        """
        if self._limit:
            return self._limit.get_sql()
        return ''

    def find_table(self, table):
        """
        Finds a table by name or alias. The FROM tables and JOIN tables
        are included in the search.

        :type table: str or :class:`ModelBase <django:django.db.models.base.ModelBase>`
        :param table: string of the table name or alias or a ModelBase instance

        :return: The table if it is found, otherwise None
        :rtype: Table or None
        """
        table = TableFactory(table)
        identifier = table.get_identifier()
        join_tables = [join_item.right_table for join_item in self.joins]
        for table in (self.tables + join_tables):
            if table.get_identifier() == identifier:
                return table
        return None

    # TODO: add test for alias
    # TODO: add option to use explicit names or *
    # TODO: add test for optional explicit names
    def wrap(self, alias=None):
        """
        Wraps the query by selecting all fields from itself

        :rtype: :class:`Query <querybuilder.query.Query>`
        :return: The wrapped query
        """
        field_names = self.get_field_names()
        query = Query(self.connection).from_table(deepcopy(self), alias=alias)
        self.__dict__.update(query.__dict__)

        # set explicit field names
        self.tables[0].set_fields(field_names)
        field_names = self.get_field_names()

        return self

    def copy(self):
        """
        Deeply copies everything in the query object except the connection object is shared
        """
        connection = self.connection
        del self.connection
        copied_query = deepcopy(self)
        copied_query.connection = connection
        self.connection = connection
        return copied_query

    def get_args(self):
        """
        Gets the args for the query which will be escaped when being executed by the
        db. All inner queries are inspected and their args are combined with this
        query's args.

        :return: all args for this query as a dict
        :rtype: dict
        """
        for table in self.tables + self.with_tables:
            if type(table) is QueryTable:
                self._where.args.update(table.query.get_args())

        return self._where.args

    def explain(self, sql=None, sql_args=None):
        """
        Runs EXPLAIN on this query

        :type sql: str or None
        :param sql: The sql to run EXPLAIN on. If None is specified, the query will
            use ``self.get_sql()``

        :type sql_args: dict or None
        :param sql_args: A dictionary of the arguments to be escaped in the query. If None and
            sql is None, the query will use ``self.get_args()``

        :rtype: list of str
        :return: list of each line of output from the EXPLAIN statement
        """
        cursor = self.get_cursor()
        if sql is None:
            sql = self.get_sql()
            sql_args = self.get_args()
        elif sql_args is None:
            sql_args = {}

        cursor.execute('EXPLAIN {0}'.format(sql), sql_args)
        rows = self._fetch_all_as_dict(cursor)
        return rows

    def select(self, return_models=False, nest=False, bypass_safe_limit=False, sql=None, sql_args=None):
        """
        Executes the SELECT statement and returns the rows as a list of dictionaries or a list of
        model instances

        :type return_models: bool
        :param return_models: Set to True to return a list of models instead of a list of dictionaries.
            Defaults to False

        :type nest: bool
        :param nest: Set to True to treat all double underscores in keynames as nested data. This will
            convert all keys with double underscores to dictionaries keyed off of the left side of
            the underscores. Ex: {"id": 1", "account__id": 1, "account__name": "Name"} becomes
            {"id": 1, "account": {"id": 1, "name": "Name"}}

        :type bypass_safe_limit: bool
        :param bypass_safe_limit: Ignores the safe_limit option even if the safe_limit is enabled

        :type sql: str or None
        :param sql: The sql to execute in the SELECT statement. If one is not specified, then the
            query will use ``self.get_sql()``

        :type sql_args: str or None
        :param sql_args: The sql args to be used in the SELECT statement. If none are specified, then
            the query wil use ``self.get_args()``

        :rtype: list of dict
        :return: list of dictionaries of the rows
        """
        # Check if we need to set a safe limit
        if bypass_safe_limit is False:
            if Query.enable_safe_limit:
                if self.count() > Query.safe_limit:
                    self.limit(Query.safe_limit)

        # determine which sql to use
        if sql is None:
            sql = self.get_sql()

        # determine which sql args to use
        if sql_args is None:
            sql_args = self.get_args()

        # get the cursor to execute the query
        cursor = self.get_cursor()

        # execute the query
        cursor.execute(sql, sql_args)

        # get the results as a list of dictionaries
        rows = self._fetch_all_as_dict(cursor)

        # check if models should be returned instead of dictionaries
        if return_models:

            # set nesting to true, so the nested models can easily load the data
            nest = True

            # build model map of map name to model
            model_map = {}
            for join_item in self.joins:
                model_map[join_item.right_table.field_prefix] = join_item.right_table.model

        # check if results should be nested
        if nest:

            # convert keys with double underscores to dictionaries
            for row in rows:
                _row = row.copy()
                for key, value in _row.items():
                    set_value_for_keypath(row, key, value, True, '__')
                    if '__' in key:
                        row.pop(key)

            # create models if needed
            if return_models:
                model_class = self.tables[0].model
                new_rows = []
                for row in rows:
                    model = model_class()
                    # assign all non-model keys first because django 1.5 requires
                    # that the model has an id set before setting a property that is
                    # a foreign key
                    for key, value in row.items():
                        if key not in model_map:
                            setattr(model, key, value)
                    # assign all model instances
                    for key, value in row.items():
                        if key in model_map:
                            child_model = model_map[key]()
                            for child_key, child_value in value.items():
                                setattr(child_model, child_key, child_value)
                            value = child_model
                        setattr(model, key, value)
                    new_rows.append(model)
                rows = new_rows

        return rows

    def insert(self, rows):
        """
        Inserts records into the db
        # TODO: implement this
        """
        if len(rows) == 0:
            return

        sql, sql_args = self.get_insert_sql(rows)

        # get the cursor to execute the query
        cursor = self.get_cursor()

        # execute the query
        cursor.execute(sql, sql_args)

    def update(self, rows):
        """
        Updates records in the db
        # TODO: implement this
        """
        if len(rows) == 0:
            return

        sql, sql_args = self.get_update_sql(rows)

        # get the cursor to execute the query
        cursor = self.get_cursor()

        # execute the query
        cursor.execute(sql, sql_args)

    def sql_delete(self):
        """
        Deletes records from the db
        # TODO: implement this
        """
        pass

    def get_count_query(self):
        """
        Copies the query object and alters the field list and order by to do a more efficient count
        """
        query_copy = self.copy()
        if not query_copy.tables:
            raise Exception('No tables specified to do a count')

        for table in query_copy.tables:
            del table.fields[:]

        query_copy.tables[0].add_field(CountField('*'))
        del query_copy.sorters[:]
        return query_copy

    def count(self, field='*'):
        """
        Returns a COUNT of the query by wrapping the query and performing a COUNT
        aggregate of the specified field

        :param field: the field to pass to the COUNT aggregate. Defaults to '*'
        :type field: str

        :return: The number of rows that the query will return
        :rtype: int
        """
        rows = self.get_count_query().select(bypass_safe_limit=True)
        return list(rows[0].values())[0]

    def max(self, field):
        """
        Returns the maximum value of a field in the result set of the query
        by wrapping the query and performing a MAX aggregate of the specified field
        :param field: the field to pass to the MAX aggregate
        :type field: str

        :return: The maximum value of the specified field
        :rtype: int
        """
        q = Query(self.connection).from_table(self, fields=[
            MaxField(field)
        ])
        rows = q.select(bypass_safe_limit=True)
        return list(rows[0].values())[0]

    def min(self, field):
        """
        Returns the minimum value of a field in the result set of the query
        by wrapping the query and performing a MIN aggregate of the specified field
        :param field: the field to pass to the MIN aggregate
        :type field: str

        :return: The minimum value of the specified field
        :rtype: int
        """
        q = Query(self.connection).from_table(self, fields=[
            MinField(field)
        ])
        rows = q.select(bypass_safe_limit=True)
        return list(rows[0].values())[0]

    def sum(self, field):
        """
        Returns the sum of the field in the result set of the query
        by wrapping the query and performing a SUM aggregate of the specified field
        :param field: the field to pass to the SUM aggregate
        :type field: str

        :return: The sum of the specified field
        :rtype: int
        """
        q = Query(self.connection).from_table(self, fields=[
            SumField(field)
        ])
        rows = q.select(bypass_safe_limit=True)
        return list(rows[0].values())[0]

    def avg(self, field):
        """
        Returns the average of the field in the result set of the query
        by wrapping the query and performing an AVG aggregate of the specified field
        :param field: the field to pass to the AVG aggregate
        :type field: str

        :return: The average of the specified field
        :rtype: int
        """
        q = Query(self.connection).from_table(self, fields=[
            AvgField(field)
        ])
        rows = q.select(bypass_safe_limit=True)
        return list(rows[0].values())[0]

    def _fetch_all_as_dict(self, cursor):
        """
        Iterates over the result set and converts each row to a dictionary

        :return: A list of dictionaries where each row is a dictionary
        :rtype: list of dict
        """
        desc = cursor.description
        return [
            dict(zip([col[0] for col in desc], row))
            for row in cursor.fetchall()
        ]


class QueryWindow(Query):
    """
    This is a query window that is meant to be used in the OVER clause of
    window functions. It extends ``Query``, but the only methods that will
    be used are ``order_by`` and ``partition_by`` (which just calls ``group_by``)
    """

    def partition_by(self, field=None, table=None):
        """
        Equivalent to ``order_by``, but named accordingly to the syntax of
        a window function

        :type field: str or dict or Field
        :param field: This can be a string of a field name, a dict of {'alias': field}, or
            a ``Field`` instance

        :type table: str or dict or Table
        :param table: Optional. This can be a string of a table name, a dict of {'alias': table}, or
            a ``Table`` instance. A table only needs to be supplied in more complex queries where
            the field name is ambiguous.

        :rtype: :class:`querybuilder.query.QueryWindow`
        :return: self
        """
        return self.group_by(field, table)

    def get_sql(self, debug=False, use_cache=True):
        """
        Generates the sql for this query window and returns the sql as a string.

        :type debug: bool
        :param debug: If True, the sql will be returned in a format that is easier to read and debug.
            Defaults to False

        :type use_cache: bool
        :param use_cache: If True, the query will returned the cached sql if it exists rather
            then generating the sql again. If False, the sql will be generated again. Defaults to True.

        :rtype: str
        :return: The generated sql for this query window
        """
        # TODO: implement caching and debug
        sql = ''
        sql += self.build_partition_by_fields()
        sql += self.build_order_by(use_alias=False)
        sql += self.build_limit()
        sql = sql.strip()
        sql = 'OVER ({0})'.format(sql)
        self.sql = sql

        return self.sql

    def build_partition_by_fields(self):
        """
        Equivalent to ``self.build_groups()`` except for the GROUP BY
        clause being named PARTITION BY

        :return: The sql to be used in the PARTITION BY clause
        :rtype: str
        """
        select_sql = self.build_groups()
        return select_sql.replace('GROUP BY', 'PARTITION BY', 1)


class QueryBuilderQuerySet(QuerySet):

    class Meta:
        model = None

    def __init__(self, model=None, query=None, using=None, **kwargs):
        if self.Meta is not None and model is None and hasattr(self.Meta, 'model'):
            model = self.Meta.model
            if isinstance(model, str):
                model = get_model(*model.split('.', 1))
        super(QueryBuilderQuerySet, self).__init__(model, query, using, **kwargs)
        self._queryset = self.model.objects.get_queryset()

    def __getitem__(self, k):
        if isinstance(k, int):
            return self.get_model_queryset(self._queryset, k, 1)[0]
        elif hasattr(k, 'start'):
            return self.get_model_queryset(
                self._queryset,
                k.start,
                k.stop
            )
        return None

    def get_model_queryset(self, queryset, offset, limit):
        raise NotImplementedError

    def get_field_name_from_filter(self, filter):
        filter_bits = filter.split(LOOKUP_SEP)
        field_name = filter_bits.pop(0)
        return field_name

    def call_field_filter_method(self, field, value, type='filter'):
        field_name = self.get_field_name_from_filter(field)
        filter_method_name = '{0}__{1}'.format(
            type,
            field_name
        )
        default_filter_method_name = '{0}__'.format(
            type
        )
        filter_method = getattr(self, default_filter_method_name)
        if hasattr(self, filter_method_name) and value is not None:
            filter_method = getattr(self, filter_method_name)
        filter_method({field: value}, field, value)

    def filter__(self, filter, field, value):
        pass

    def filter(self, *args, **kwargs):
        for field, value in six.iteritems(kwargs):
            self.call_field_filter_method(field, value, type='filter')
        return self

    def exclude__(self, filter, field, value):
        pass

    def exclude(self, *args, **kwargs):
        for field, value in six.iteritems(kwargs):
            self.call_field_filter_method(field, value, type='exclude')
        return self

    def count(self):
        raise NotImplementedError

    def order__(self, field, desc=False):
        pass

    def order_by(self, *field_names):
        for field in field_names:
            if field == 'pk':
                field = self.model._meta.pk.name
            desc = False
            if field[0] == '-':
                field = field[1:]
                desc = True
            method_name = '{0}__{1}'.format(
                'order',
                field
            )
            default_method_name = '{0}__'.format(
                'order'
            )
            method = getattr(self, default_method_name)
            if hasattr(self, method_name):
                method = getattr(self, method_name)
            method(field, desc)
        return self

    def distinct(self, *field_names):
        raise NotImplementedError


class JsonQueryset(QueryBuilderQuerySet):

    def __init__(self, *args, **kwargs):
        super(JsonQueryset, self).__init__(*args, **kwargs)
        self.json_query = Query().from_table(self.model)

    def get_model_queryset(self, queryset, offset, limit):
        return [self.model(**fields) for fields in self.json_query.limit(limit, offset).select()]

    def count(self):
        return self.json_query.count()

    def order_by(self, *field_names):
        for field_name in field_names:
            if field_name == 'pk':
                field_name = self.model._meta.pk.name
            reverse = '-' if field_name[0] == '-' else ''
            field_name = field_name.lstrip('-')
            parts = field_name.split('->')
            if len(parts) == 2:
                self.json_query.order_by('{0}{1}->>\'{2}\''.format(reverse, parts[0], parts[1]))
            else:
                self.json_query.order_by('{0}{1}'.format(reverse, field_name))
        return self

    def filter(self, *args, **kwargs):
        for key, value in kwargs.items():
            key = key.replace('__exact', '')
            parts = key.split('->')
            if len(parts) == 2:
                key = '{0}->>\'{1}\''.format(parts[0], parts[1])
                value = six.u('{0}'.format(value))
            if hasattr(value, 'id'):
                key = '{0}_id'.format(key)
                self.json_query.where(**{'{0}'.format(key): value.id})
            else:
                self.json_query.where(**{key: value})
        return self

    def limit(self, *args, **kwargs):
        self.json_query.limit(*args, **kwargs)
        return self



from django.db import connection


class LogManager(object):

    loggers = {}
    debug = False

    @staticmethod
    def reset():
        LogManager.loggers = {}
        LogManager.debug = False
        connection.queries[:] = []

    @staticmethod
    def add_logger(logger):
        LogManager.enable_logging()
        LogManager.loggers[logger.name] = logger

    @staticmethod
    def get_logger(name=None):
        if name in LogManager.loggers:
            return LogManager.loggers[name]
        return Logger(name)

    @staticmethod
    def enable_logging():
        LogManager.debug = True

    @staticmethod
    def disable_logging():
        LogManager.debug = False


class Logger(object):

    def __init__(self, name=None):
        if name is None:
            name = 'default'
        self.name = name
        self.query_index = None
        self.queries = []
        LogManager.add_logger(self)

    def start_logging(self):
        self.query_index = len(connection.queries)

    def update_log(self):
        num_queries = len(connection.queries)
        if self.query_index is not None and num_queries > self.query_index and LogManager.debug:
            self.queries.extend(connection.queries[self.query_index:])
        self.query_index = num_queries

    def get_log(self):
        self.update_log()
        return self.queries

    def stop_logging(self):
        self.update_log()
        self.query_index = None

    def clear_log(self):
        self.queries = []
        if self.query_index is not None:
            self.query_index = len(connection.queries)

    def count(self):
        self.update_log()
        return len(self.queries)



from django.test import TestCase
from querybuilder.fields import Field
from querybuilder.query import Query
from querybuilder.tables import TableFactory, SimpleTable, ModelTable, QueryTable
from querybuilder.tests.models import Account


class TableFactoryTest(TestCase):
    """
    Tests the functionality of the table factory
    """
    def test_str(self):
        """
        Verifies that a table is generated for a str object and that kwargs are passed
        """
        table = TableFactory('test', fields=['one'])
        self.assertIsInstance(table, SimpleTable)
        self.assertEqual(1, len(table.fields))

    def test_unicode(self):
        """
        Verifies that a table is generated for a unicode object and that kwargs are passed
        """
        table = TableFactory(u'test', fields=['one'])
        self.assertIsInstance(table, SimpleTable)
        self.assertEqual(1, len(table.fields))

    def test_model(self):
        """
        Verifies that a table is generated from a model and that kwargs are passed
        """
        table = TableFactory(Account, fields=['one'])
        self.assertIsInstance(table, ModelTable)
        self.assertEqual(1, len(table.fields))

    def test_query(self):
        """
        Verifies that a table is generated from a query object and that kwargs are passed
        """
        table = TableFactory(Query().from_table('test'), fields=['one'])
        self.assertIsInstance(table, QueryTable)
        self.assertEqual(1, len(table.fields))

    def test_table(self):
        """
        Verifies that kwargs are properly set from another table object and that kwargs are passed
        """
        test_table = TableFactory('test', alias='test_alias')
        table = TableFactory(test_table, fields=['one'])
        self.assertIsInstance(table, SimpleTable)
        self.assertEqual(1, len(table.fields))
        self.assertEqual('test_alias', table.alias)

    def test_none(self):
        """
        Verifies that None is returned for unsupported data
        """
        table = TableFactory(5, fields=['one'])
        self.assertIsNone(table)

    def test_dict(self):
        """
        Verifies that dict alias data is properly extracted
        """
        table = TableFactory({'test_alias': Account}, fields=['one'])
        self.assertIsInstance(table, ModelTable)
        self.assertEqual(1, len(table.fields))
        self.assertEqual('test_alias', table.alias)


class TableTest(TestCase):
    """
    Tests functionality of the Table class
    """
    def test_add_fields_str(self):
        """
        Tests calling add_fields with a str object
        """
        table = TableFactory('test')
        fields = table.add_fields('one')
        self.assertEqual(1, len(fields))
        self.assertIsNotNone(fields[0])
        self.assertEqual('test.one', fields[0].get_identifier())

    def test_add_fields_unicode(self):
        """
        Tests calling add_fields with a unicode object
        """
        table = TableFactory('test')
        fields = table.add_fields(u'one')
        self.assertEqual(1, len(fields))
        self.assertIsNotNone(fields[0])
        self.assertEqual('test.one', fields[0].get_identifier())

    def test_add_fields_tuple(self):
        """
        Tests calling add_fields with a tuple of names
        """
        table = TableFactory('test')
        fields = table.add_fields(('one', 'two'))
        self.assertEqual(2, len(fields))
        self.assertIsNotNone(fields[0])
        self.assertIsNotNone(fields[1])
        self.assertEqual('test.one', fields[0].get_identifier())
        self.assertEqual('test.two', fields[1].get_identifier())

    def test_add_fields_list(self):
        """
        Tests calling add_fields with a list of names
        """
        table = TableFactory('test')
        fields = table.add_fields(['one', 'two'])
        self.assertEqual(2, len(fields))
        self.assertIsNotNone(fields[0])
        self.assertIsNotNone(fields[1])
        self.assertEqual('test.one', fields[0].get_identifier())
        self.assertEqual('test.two', fields[1].get_identifier())

    def test_add_existing_field(self):
        """
        Tests the add field function to make sure an existing field is ignored
        """
        table = TableFactory('test')
        field = table.add_field('one')
        self.assertIsInstance(field, Field)
        field = table.add_field('one')
        self.assertIsNone(field)

    def test_remove_field(self):
        """
        Verifies that a field is properly removed
        """
        table = TableFactory('test')
        table.add_fields(['one', 'two'])
        self.assertEqual(2, len(table.fields))
        field = table.remove_field('one')
        self.assertIsNotNone(field)
        self.assertEqual('test.one', field.get_identifier())
        self.assertEqual(1, len(table.fields))
        field = table.remove_field('one')
        self.assertIsNone(field)

    def test_get_field_identifiers(self):
        """
        Makes sure correct identifiers are returned for all fields
        """
        table = TableFactory('test')
        table.add_fields(['one', 'two'])
        self.assertEqual(['test.one', 'test.two'], table.get_field_identifiers())

    def test_find_field(self):
        """
        Verifies that the correct field is returned
        """
        table = TableFactory('test')
        table.add_fields(['one', 'two'])
        field = table.find_field('one')
        self.assertEqual('test.one', field.get_identifier())
        self.assertEqual(table.fields[0], field)

    def test_find_field_alias(self):
        """
        Verifies that the correct field is returned when passing an alias
        """
        table = TableFactory('test')
        table.add_fields(['one', {'my_alias': 'two'}])
        field = table.find_field(alias='my_alias')
        self.assertEqual('my_alias', field.get_identifier())
        self.assertEqual(table.fields[1], field)

    def test_find_field_none(self):
        """
        Verifies that None is returned when no field is found
        """
        table = TableFactory('test')
        table.add_fields(['one', 'two'])
        field = table.find_field('fake_field')
        self.assertIsNone(field)


class QueryTableTest(TestCase):
    """
    Tests the QueryTable object
    """
    def test_get_from_name(self):
        """
        Verifies that the correct name is generated
        """
        inner_query = Query().from_table(Account)
        query = Query().from_table(inner_query)
        self.assertEqual('(SELECT tests_account.* FROM tests_account)', query.tables[0].get_from_name())



def value_for_keypath(dict, keypath):
    """
    Returns the value of a keypath in a dictionary
    if the keypath exists or None if the keypath
    does not exist.
    """

    if len(keypath) == 0:
        return dict

    keys = keypath.split('.')
    value = dict
    for key in keys:
        if key in value:
            value = value[key]
        else:
            return None

    return value


def set_value_for_keypath(item, keypath, value, create_if_needed=False, delimeter='.'):
    """
    Sets the value for a keypath in a dictionary
    if the keypath exists. This modifies the
    original dictionary.
    """

    if len(keypath) == 0:
        return None

    keys = keypath.split(delimeter)
    if len(keys) > 1:
        key = keys[0]

        if create_if_needed:
            item[key] = item.get(key, {})

        if key in item:
            if set_value_for_keypath(item[key], delimeter.join(keys[1:]), value,
                                     create_if_needed=create_if_needed, delimeter=delimeter):
                return item

        return None

    if create_if_needed:
        item[keypath] = item.get(keypath, {})

    if keypath in item:
        item[keypath] = value
        return item
    else:
        return None



from querybuilder.tests.models import Account, Order, User
from querybuilder.query import Query
from querybuilder.tests.query_tests import QueryTestCase, get_comparison_str


class JoinTest(QueryTestCase):

    def test_join_str_to_str(self):
        query = Query().from_table(
            table='test_table'
        ).join(
            'other_table',
            condition='other_table.test_id = test_table.id'
        )

        query_str = query.get_sql()
        expected_query = (
            'SELECT test_table.* '
            'FROM test_table '
            'JOIN other_table ON other_table.test_id = test_table.id'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_join_model_to_str(self):
        query = Query().from_table(
            table=Account
        ).join(
            'other_table',
            condition='other_table.test_id = tests_account.id'
        )

        query_str = query.get_sql()
        expected_query = (
            'SELECT tests_account.* '
            'FROM tests_account '
            'JOIN other_table ON other_table.test_id = tests_account.id'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_no_fields_and_fields(self):
        query = Query().from_table(
            table=Account,
            fields=None
        ).join(
            'other_table',
            fields=[
                'field_one',
                'field_two'
            ],
            prefix_fields=True,
            condition='other_table.test_id = tests_account.id'
        )

        query_str = query.get_sql()
        expected_query = (
            'SELECT other_table.field_one AS other_table__field_one, '
            'other_table.field_two AS other_table__field_two '
            'FROM tests_account '
            'JOIN other_table ON other_table.test_id = tests_account.id'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    # TODO: the foreign key join should do an extra query and
    # merge the results in python afterwards
    def test_join_model_foreign_key(self):
        query = Query().from_table(
            table=Account
        ).join(
            Order,
        )

        query_str = query.get_sql()
        expected_query = (
            'SELECT tests_account.* '
            'FROM tests_account '
            'JOIN tests_order ON tests_order.account_id = tests_account.id'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_join_model_foreign_key_reverse(self):
        query = Query().from_table(
            table=Order
        ).join(
            Account,
        )

        query_str = query.get_sql()
        expected_query = (
            'SELECT tests_order.* '
            'FROM tests_order '
            'JOIN tests_account ON tests_account.id = tests_order.account_id'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_join_model_one_to_one(self):
        query = Query().from_table(
            table=Account
        ).join(
            User,
        )

        query_str = query.get_sql()
        expected_query = (
            'SELECT tests_account.* '
            'FROM tests_account '
            'JOIN tests_user ON tests_user.id = tests_account.user_id'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_join_model_one_to_one_reverse(self):
        query = Query().from_table(
            table=User
        ).join(
            Account,
        )

        query_str = query.get_sql()
        expected_query = (
            'SELECT tests_user.* '
            'FROM tests_user '
            'JOIN tests_account ON tests_account.user_id = tests_user.id'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_join_model_fields(self):
        query = Query().from_table(
            table=Account,
            fields=[
                'one',
                'two',
            ]
        ).join(
            Order,
            fields=[{
                'three': 'one'
            }, {
                'four': 'two'
            }]
        )

        query_str = query.get_sql()
        expected_query = (
            'SELECT tests_account.one, '
            'tests_account.two, '
            'tests_order.one AS three, '
            'tests_order.two AS four '
            'FROM tests_account '
            'JOIN tests_order ON tests_order.account_id = tests_account.id'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_join_model_fields_all(self):
        query = Query().from_table(
            table=Account,
            fields=[
                '*',
            ]
        ).join(
            Order,
            fields=[
                '*'
            ],
            extract_fields=False
        )

        query_str = query.get_sql()
        expected_query = (
            'SELECT tests_account.*, '
            'tests_order.* '
            'FROM tests_account '
            'JOIN tests_order ON tests_order.account_id = tests_account.id'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_join_model_fields_extract(self):
        query = Query().from_table(
            table=Account,
            fields=[
                '*',
            ]
        ).join(
            Order,
            fields=[
                '*'
            ]
        )

        query_str = query.get_sql()
        expected_query = (
            'SELECT tests_account.*, '
            'tests_order.id, '
            'tests_order.account_id, '
            'tests_order.revenue, '
            'tests_order.margin, '
            'tests_order.margin_percent, '
            'tests_order.time '
            'FROM tests_account '
            'JOIN tests_order ON tests_order.account_id = tests_account.id'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_join_model_fields_prefix(self):
        query = Query().from_table(
            table=Account,
            fields=[
                '*',
            ]
        ).join(
            Order,
            fields=[
                'id',
                'margin',
            ],
            prefix_fields=True
        )

        query_str = query.get_sql()

        expected_query = (
            'SELECT tests_account.*, '
            'tests_order.id AS order__id, '
            'tests_order.margin AS order__margin '
            'FROM tests_account '
            'JOIN tests_order ON tests_order.account_id = tests_account.id'
        )
        self.assertEqual(query_str, expected_query)



from django.test.utils import override_settings

from querybuilder.logger import Logger, LogManager
from querybuilder.query import Query
from querybuilder.tests.models import Account
from querybuilder.tests.query_tests import QueryTestCase


@override_settings(DEBUG=True)
class InsertTest(QueryTestCase):

    def setUp(self):
        self.logger = Logger()
        self.logger.start_logging()

    def tearDown(self):
        super(InsertTest, self).tearDown()
        LogManager.loggers = {}

    def test_insert_single_row(self):
        query = Query().from_table(
            table=Account,
            fields=[
                'user_id',
                'first_name',
                'last_name'
            ]
        )

        rows = [
            [1, 'Test', 'User']
        ]

        sql, sql_params = query.get_insert_sql(rows)

        self.assertEqual(
            sql,
            (
                'INSERT INTO tests_account (user_id, first_name, last_name) VALUES (%s, %s, %s)'
            )
        )
        self.assertEqual(sql_params[0], 1)
        self.assertEqual(sql_params[1], 'Test')
        self.assertEqual(sql_params[2], 'User')

        query.insert(rows)
        sql = self.logger.get_log()[0]['sql']
        self.assertEqual(sql, "INSERT INTO tests_account (user_id, first_name, last_name) VALUES (1, 'Test', 'User')")

    def test_insert_multiple_rows(self):
        query = Query().from_table(
            table=Account,
            fields=[
                'user_id',
                'first_name',
                'last_name'
            ]
        )

        rows = [
            [1, 'Test', 'User'],
            [2, 'Test2', 'User2'],
        ]

        sql, sql_params = query.get_insert_sql(rows)

        self.assertEqual(
            sql,
            'INSERT INTO tests_account (user_id, first_name, last_name) VALUES (%s, %s, %s), (%s, %s, %s)'
        )
        self.assertEqual(sql_params[0], 1)
        self.assertEqual(sql_params[1], 'Test')
        self.assertEqual(sql_params[2], 'User')
        self.assertEqual(sql_params[3], 2)
        self.assertEqual(sql_params[4], 'Test2')
        self.assertEqual(sql_params[5], 'User2')

        query.insert(rows)
        sql = self.logger.get_log()[0]['sql']
        self.assertEqual(
            sql,
            ("INSERT INTO tests_account (user_id, first_name, last_name) "
             "VALUES (1, 'Test', 'User'), (2, 'Test2', 'User2')")
        )



from django.db import connection
from django.test import TestCase
from django.test.utils import override_settings

from querybuilder.logger import Logger, LogManager
from querybuilder.query import Query
from querybuilder.tests.models import Account


@override_settings(DEBUG=True)
class LogManagerTest(TestCase):
    """
    Includes functions to test the LogManager
    """
    def test_log_manager(self):
        self.assertEqual(len(LogManager.loggers.items()), 0, 'Incorrect number of loggers')
        logger_one = LogManager.get_logger('one')
        self.assertEqual(len(LogManager.loggers), 1, 'Incorrect number of loggers')
        logger_one = LogManager.get_logger('one')
        self.assertEqual(len(LogManager.loggers), 1, 'Incorrect number of loggers')
        LogManager.get_logger('two')
        self.assertEqual(len(LogManager.loggers), 2, 'Incorrect number of loggers')

        logger_one.start_logging()
        query = Query().from_table(Account)
        query.select()

        self.assertEqual(logger_one.count(), 1, 'Incorrect number of queries')
        LogManager.disable_logging()
        query.select()
        self.assertEqual(logger_one.count(), 1, 'Incorrect number of queries')
        LogManager.enable_logging()
        query.select()
        self.assertEqual(logger_one.count(), 2, 'Incorrect number of queries')


@override_settings(DEBUG=True)
class LoggerTest(TestCase):
    """
    Includes functions to test the Logger
    """
    def setUp(self):
        super(LoggerTest, self).setUp()
        LogManager.reset()

    def test_init(self):
        """
        Tests the init method
        """
        logger = Logger()
        self.assertEqual('default', logger.name)
        self.assertIsNone(logger.query_index)
        self.assertEqual(0, len(logger.queries))
        self.assertEqual(1, len(LogManager.loggers))

        logger = Logger('custom_name')
        self.assertEqual('custom_name', logger.name)
        self.assertEqual(2, len(LogManager.loggers))

    def test_start_logging(self):
        """
        Verifies that the query index gets updated
        """
        logger = Logger()

        query = Query().from_table(Account)
        query.select()
        query.select()
        logger.start_logging()
        self.assertEqual(logger.query_index, len(connection.queries))

    def test_count(self):
        """
        Verifies that the correct number of queries is returned
        """
        logger = Logger()
        logger.start_logging()

        query = Query().from_table(Account)
        query.select()
        query.select()
        self.assertEqual(2, logger.count())
        query.select()
        self.assertEqual(3, logger.count())

    def test_stop_logging(self):
        """
        Verifies that the logger stops caring about queries
        """
        logger = Logger()
        logger.start_logging()

        query = Query().from_table(Account)
        query.select()
        query.select()

        self.assertEqual(2, logger.count())

        logger.stop_logging()
        query.select()
        query.select()
        self.assertEqual(2, logger.count())

        logger.start_logging()
        query.select()
        self.assertEqual(3, logger.count())

    def test_get_log(self):
        """
        Verifies that queries get returned
        """
        pass

    def test_update_log(self):
        """
        Verifies that the log gets updated properly
        """
        pass

    def test_logger(self):
        logger_one = Logger('one')
        logger_two = Logger('two')

        logger_one.start_logging()
        query = Query().from_table(Account)
        query.select()

        self.assertEqual(logger_one.count(), 1, 'Incorrect number of queries')

        query.select()
        logger_two.start_logging()
        query.select()
        logger_one.stop_logging()
        query.select()

        self.assertEqual(logger_one.count(), 3, 'Incorrect number of queries')
        self.assertEqual(logger_two.count(), 2, 'Incorrect number of queries')

        query.select()
        logger_one.start_logging()
        query.select()

        self.assertEqual(logger_one.count(), 4, 'Incorrect number of queries')
        self.assertEqual(logger_two.count(), 4, 'Incorrect number of queries')

        query.select()
        logger_two.clear_log()
        query.select()

        self.assertEqual(logger_one.count(), 6, 'Incorrect number of queries')
        self.assertEqual(logger_two.count(), 1, 'Incorrect number of queries')

    def test_clear_log(self):
        """
        Makes sure queries are cleared
        """
        logger_one = Logger('one')
        logger_one.start_logging()
        query = Query().from_table(Account)

        # run a query and update the logger's query list
        query.select()
        logger_one.update_log()

        # there should be one query
        self.assertEqual(logger_one.count(), 1)

        # increment the connection query count
        query.select()

        # clear the log
        logger_one.clear_log()

        # make sure no queries
        self.assertEqual(0, len(logger_one.queries))

    def test_clear_log_no_index(self):
        """
        Makes sure that the query index doesn't change
        """
        logger_one = Logger('one')

        query = Query().from_table(Account)
        query.select()

        self.assertIsNone(logger_one.query_index)

        # clear the log
        logger_one.clear_log()

        # make sure no query index
        self.assertIsNone(logger_one.query_index)



from querybuilder.fields import Year, Month, Hour, Minute, Second, NoneTime, AllTime
from querybuilder.query import Query
from querybuilder.tests.models import Order
from querybuilder.tests.query_tests import QueryTestCase, get_comparison_str


class DateTest(QueryTestCase):
    def test_year(self):
        query = Query().from_table(
            table=Order,
            fields=[
                Year('time')
            ]
        )
        query_str = query.get_sql()
        expected_query = 'SELECT CAST(EXTRACT(year FROM tests_order.time) AS INT) AS time__year FROM tests_order'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_year_auto(self):
        query = Query().from_table(
            table=Order,
            fields=[
                Year('time', auto=True)
            ]
        )
        query_str = query.get_sql()
        expected_query = (
            'SELECT CAST(EXTRACT(year FROM tests_order.time) AS INT) AS time__year, '
            'CAST(EXTRACT(epoch FROM date_trunc(\'year\', tests_order.time)) AS INT) AS time__epoch '
            'FROM tests_order '
            'GROUP BY time__year, time__epoch '
            'ORDER BY time__epoch ASC'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_year_auto_desc(self):
        query = Query().from_table(
            table=Order,
            fields=[
                Year('time', auto=True, desc=True)
            ]
        )
        query_str = query.get_sql()
        expected_query = (
            'SELECT CAST(EXTRACT(year FROM tests_order.time) AS INT) AS time__year, '
            'CAST(EXTRACT(epoch FROM date_trunc(\'year\', tests_order.time)) AS INT) AS time__epoch '
            'FROM tests_order '
            'GROUP BY time__year, time__epoch '
            'ORDER BY time__epoch DESC'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_month_auto(self):
        query = Query().from_table(
            table=Order,
            fields=[
                Month('time', auto=True)
            ]
        )
        query_str = query.get_sql()
        expected_query = (
            'SELECT CAST(EXTRACT(year FROM tests_order.time) AS INT) AS time__year, '
            'CAST(EXTRACT(month FROM tests_order.time) AS INT) AS time__month, '
            'CAST(EXTRACT(epoch FROM date_trunc(\'month\', tests_order.time)) AS INT) AS time__epoch '
            'FROM tests_order '
            'GROUP BY time__year, time__month, time__epoch '
            'ORDER BY time__epoch ASC'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_hour_auto(self):
        query = Query().from_table(
            table=Order,
            fields=[
                Hour('time', auto=True)
            ]
        )
        query_str = query.get_sql()
        expected_query = (
            'SELECT CAST(EXTRACT(year FROM tests_order.time) AS INT) AS time__year, '
            'CAST(EXTRACT(month FROM tests_order.time) AS INT) AS time__month, '
            'CAST(EXTRACT(day FROM tests_order.time) AS INT) AS time__day, '
            'CAST(EXTRACT(hour FROM tests_order.time) AS INT) AS time__hour, '
            'CAST(EXTRACT(epoch FROM date_trunc(\'hour\', tests_order.time)) AS INT) AS time__epoch '
            'FROM tests_order '
            'GROUP BY time__year, time__month, time__day, time__hour, time__epoch '
            'ORDER BY time__epoch ASC'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_minute_auto(self):
        query = Query().from_table(
            table=Order,
            fields=[
                Minute('time', auto=True)
            ]
        )
        query_str = query.get_sql()
        expected_query = (
            'SELECT CAST(EXTRACT(year FROM tests_order.time) AS INT) AS time__year, '
            'CAST(EXTRACT(month FROM tests_order.time) AS INT) AS time__month, '
            'CAST(EXTRACT(day FROM tests_order.time) AS INT) AS time__day, '
            'CAST(EXTRACT(hour FROM tests_order.time) AS INT) AS time__hour, '
            'CAST(EXTRACT(minute FROM tests_order.time) AS INT) AS time__minute, '
            'CAST(EXTRACT(epoch FROM date_trunc(\'minute\', tests_order.time)) AS INT) AS time__epoch '
            'FROM tests_order '
            'GROUP BY time__year, time__month, time__day, time__hour, time__minute, time__epoch '
            'ORDER BY time__epoch ASC'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_second_auto(self):
        query = Query().from_table(
            table=Order,
            fields=[
                Second('time', auto=True)
            ]
        )
        query_str = query.get_sql()
        expected_query = (
            'SELECT CAST(EXTRACT(year FROM tests_order.time) AS INT) AS time__year, '
            'CAST(EXTRACT(month FROM tests_order.time) AS INT) AS time__month, '
            'CAST(EXTRACT(day FROM tests_order.time) AS INT) AS time__day, '
            'CAST(EXTRACT(hour FROM tests_order.time) AS INT) AS time__hour, '
            'CAST(EXTRACT(minute FROM tests_order.time) AS INT) AS time__minute, '
            'CAST(EXTRACT(second FROM tests_order.time) AS INT) AS time__second, '
            'CAST(EXTRACT(epoch FROM date_trunc(\'second\', tests_order.time)) AS INT) AS time__epoch '
            'FROM tests_order '
            'GROUP BY time__year, time__month, time__day, time__hour, time__minute, time__second, time__epoch '
            'ORDER BY time__epoch ASC'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_group_none(self):
        query = Query().from_table(
            table=Order,
            fields=[
                NoneTime('time')
            ]
        )
        query_str = query.get_sql()
        expected_query = (
            'SELECT CAST(EXTRACT(epoch FROM tests_order.time) AS INT) AS time__epoch '
            'FROM tests_order '
            'GROUP BY time__epoch '
            'ORDER BY time__epoch ASC'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_group_all(self):
        query = Query().from_table(
            table=Order,
            fields=[
                AllTime('time')
            ]
        )
        query_str = query.get_sql()
        expected_query = 'SELECT CAST(0 AS INT) AS time__epoch FROM tests_order'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))



from django.db import models
from jsonfield import JSONField


class User(models.Model):
    """
    User model
    """
    email = models.CharField(max_length=256)


class Account(models.Model):
    """
    Account model
    """
    user = models.OneToOneField(User)
    first_name = models.CharField(max_length=64)
    last_name = models.CharField(max_length=64)


class Order(models.Model):
    """
    Order model
    """
    account = models.ForeignKey(Account)
    revenue = models.FloatField(null=True)
    margin = models.FloatField()
    margin_percent = models.FloatField()
    time = models.DateTimeField()


class MetricRecord(models.Model):
    """
    Example metric model
    """
    other_value = models.FloatField(default=0)
    data = JSONField()



from querybuilder.query import Query
from querybuilder.tests.query_tests import QueryTestCase, get_comparison_str


class OrderByTest(QueryTestCase):
    def test_order_by_single_asc(self):
        query = Query().from_table(
            table='test_table'
        ).order_by(
            'field_one'
        )
        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table ORDER BY field_one ASC'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_order_by_many_asc(self):
        query = Query().from_table(
            table='test_table'
        ).order_by(
            'field_one'
        ).order_by(
            'field_two'
        )
        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table ORDER BY field_one ASC, field_two ASC'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_order_by_single_desc(self):
        query = Query().from_table(
            table='test_table'
        ).order_by(
            '-field_one'
        )
        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table ORDER BY field_one DESC'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_order_by_many_desc(self):
        query = Query().from_table(
            table='test_table'
        ).order_by(
            '-field_one'
        ).order_by(
            '-field_two'
        )
        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table ORDER BY field_one DESC, field_two DESC'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))



from django.test.utils import override_settings
from django_dynamic_fixture import G

from querybuilder.logger import Logger
from querybuilder.query import Query
from querybuilder.tests.models import Account, Order
from querybuilder.tests.query_tests import QueryTestCase


@override_settings(DEBUG=True)
class TestUpdate(QueryTestCase):

    def setUp(self):
        self.logger = Logger()
        self.logger.start_logging()

    def test_update_single_row(self):
        query = Query().from_table(
            table=Account,
            fields=[
                'id',
                'user_id',
                'first_name',
                'last_name'
            ]
        )

        rows = [
            [1, 1, 'Test\'s', '"User"']
        ]

        sql, sql_params = query.get_update_sql(rows)

        self.assertEqual(
            sql,
            (
                'UPDATE tests_account '
                'SET user_id = new_values.user_id, '
                'first_name = new_values.first_name, '
                'last_name = new_values.last_name '
                'FROM (VALUES (%s, %s, %s, %s)) '
                'AS new_values (id, user_id, first_name, last_name) '
                'WHERE tests_account.id = new_values.id'
            )
        )

        self.assertEqual(sql_params[0], 1)
        self.assertEqual(sql_params[1], 1)
        self.assertEqual(sql_params[2], 'Test\'s')
        self.assertEqual(sql_params[3], '"User"')

        query.update(rows)
        sql = self.logger.get_log()[0]['sql']
        self.assertEqual(
            sql,
            (
                "UPDATE tests_account "
                "SET user_id = new_values.user_id, "
                "first_name = new_values.first_name, "
                "last_name = new_values.last_name "
                "FROM (VALUES (1, 1, 'Test''s', '\"User\"')) "
                "AS new_values (id, user_id, first_name, last_name) "
                "WHERE tests_account.id = new_values.id"
            )
        )

    def test_update_multiple_rows(self):
        query = Query().from_table(
            table=Account,
            fields=[
                'id',
                'user_id',
                'first_name',
                'last_name'
            ]
        )

        rows = [
            [1, 1, 'Test', 'User'],
            [2, 2, 'Test2', 'User2']
        ]

        sql, sql_params = query.get_update_sql(rows)

        self.assertEqual(
            sql,
            (
                'UPDATE tests_account '
                'SET user_id = new_values.user_id, '
                'first_name = new_values.first_name, '
                'last_name = new_values.last_name '
                'FROM (VALUES (%s, %s, %s, %s), (%s, %s, %s, %s)) '
                'AS new_values (id, user_id, first_name, last_name) '
                'WHERE tests_account.id = new_values.id'
            )
        )
        self.assertEqual(sql_params[0], 1)
        self.assertEqual(sql_params[1], 1)
        self.assertEqual(sql_params[2], 'Test')
        self.assertEqual(sql_params[3], 'User')
        self.assertEqual(sql_params[4], 2)
        self.assertEqual(sql_params[5], 2)
        self.assertEqual(sql_params[6], 'Test2')
        self.assertEqual(sql_params[7], 'User2')

        query.update(rows)
        sql = self.logger.get_log()[0]['sql']
        self.assertEqual(
            sql,
            (
                "UPDATE tests_account "
                "SET user_id = new_values.user_id, "
                "first_name = new_values.first_name, "
                "last_name = new_values.last_name "
                "FROM (VALUES (1, 1, 'Test', 'User'), (2, 2, 'Test2', 'User2')) "
                "AS new_values (id, user_id, first_name, last_name) "
                "WHERE tests_account.id = new_values.id"
            )
        )

    def test_update_null_numbers(self):
        """
        Verifies that null values can be bulk updated
        """
        order1 = G(Order, revenue=10)
        order2 = G(Order, revenue=20)
        order3 = G(Order, revenue=30)
        query = Query().from_table(table=Order, fields=['id', 'revenue'])

        logger = Logger()
        logger.start_logging()
        rows = [[order1.id, None], [order2.id, None], [order3.id, 3000]]
        query.update(rows)

        orders = list(Order.objects.order_by('id'))
        self.assertIsNone(orders[0].revenue)
        self.assertIsNone(orders[1].revenue)
        self.assertEqual(3000, orders[2].revenue)

    def test_update_all_nulls(self):
        """
        Verifies that the sql is modified when all values for a field are null. For whatever reason,
        postgres doesn't handle this the same when all values are null
        """
        order1 = G(Order, revenue=10, margin=5)
        order2 = G(Order, revenue=20, margin=10)
        order3 = G(Order, revenue=30, margin=15)
        query = Query().from_table(table=Order, fields=['id', 'revenue', 'margin'])

        logger = Logger()
        logger.start_logging()
        rows = [[order1.id, None, 50], [order2.id, None, 100], [order3.id, None, 150]]
        query.update(rows)

        orders = list(Order.objects.order_by('id'))
        self.assertIsNone(orders[0].revenue)
        self.assertIsNone(orders[1].revenue)
        self.assertIsNone(orders[2].revenue)



from django.db.models import Q
from django.db.models.sql import OR
from querybuilder.query import Query
from querybuilder.tests.models import Account
from querybuilder.tests.query_tests import QueryTestCase, get_comparison_str


class WhereTest(QueryTestCase):

    def test_where_eq(self):
        query = Query().from_table(
            table='test_table'
        ).where(Q(
            one='two'
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE (one = %(A0)s)'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_named_arg(self):
        query = Query().from_table(
            table='test_table'
        ).where(
            one='two'
        )

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE (one = %(A0)s)'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_many_named_arg(self):
        query = Query().from_table(
            table='test_table'
        ).where(
            one='two',
            three='four'
        )

        query_str = query.get_sql()
        expected_queries = [
            'SELECT test_table.* FROM test_table WHERE (three = %(A0)s AND one = %(A1)s)',
            'SELECT test_table.* FROM test_table WHERE (one = %(A0)s AND three = %(A1)s)'
        ]
        self.assertIn(query_str, expected_queries)

    def test_where_not_eq(self):
        query = Query().from_table(
            table='test_table'
        ).where(~Q(
            one='two'
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE ((NOT(one = %(A0)s)))'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_eq_explicit(self):
        query = Query().from_table(
            table='test_table'
        ).where(Q(
            one__eq='two'
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE (one = %(A0)s)'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_not_eq_explicit(self):
        query = Query().from_table(
            table='test_table'
        ).where(~Q(
            one__eq='two'
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE ((NOT(one = %(A0)s)))'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_gt(self):
        query = Query().from_table(
            table='test_table'
        ).where(Q(
            field_name__gt=10
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE (field_name > %(A0)s)'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_not_gt(self):
        query = Query().from_table(
            table='test_table'
        ).where(~Q(
            field_name__gt=10
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE ((NOT(field_name > %(A0)s)))'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_gte(self):
        query = Query().from_table(
            table='test_table'
        ).where(Q(
            field_name__gte=10
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE (field_name >= %(A0)s)'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_not_gte(self):
        query = Query().from_table(
            table='test_table'
        ).where(~Q(
            field_name__gte=10
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE ((NOT(field_name >= %(A0)s)))'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_lt(self):
        query = Query().from_table(
            table='test_table'
        ).where(Q(
            field_name__lt=10
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE (field_name < %(A0)s)'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_not_lt(self):
        query = Query().from_table(
            table='test_table'
        ).where(~Q(
            field_name__lt=10
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE ((NOT(field_name < %(A0)s)))'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_lte(self):
        query = Query().from_table(
            table='test_table'
        ).where(Q(
            field_name__lte=10
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE (field_name <= %(A0)s)'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_not_lte(self):
        query = Query().from_table(
            table='test_table'
        ).where(~Q(
            field_name__lte=10
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE ((NOT(field_name <= %(A0)s)))'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_in_single(self):
        query = Query().from_table(
            table=Account
        ).where(Q(
            id__in=10
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT tests_account.* FROM tests_account WHERE (id IN (%(A0)s))'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_in_csv(self):
        query = Query().from_table(
            table=Account
        ).where(Q(
            id__in='10,11,12'
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT tests_account.* FROM tests_account WHERE (id IN (%(A0)s,%(A1)s,%(A2)s))'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_in_list(self):
        query = Query().from_table(
            table=Account
        ).where(Q(
            id__in=[10, 11, 12]
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT tests_account.* FROM tests_account WHERE (id IN (%(A0)s,%(A1)s,%(A2)s))'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_contains(self):
        query = Query().from_table(
            table='test_table'
        ).where(Q(
            field_name__contains='some value'
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE (field_name LIKE %(A0)s)'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))
        self.assertEqual(query._where.args['A0'], '%some value%', 'Value is not correct')

    def test_where_not_contains(self):
        query = Query().from_table(
            table='test_table'
        ).where(~Q(
            field_name__contains='some value'
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE ((NOT(field_name LIKE %(A0)s)))'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))
        self.assertEqual(query._where.args['A0'], '%some value%', 'Value is not correct')

    def test_where_startswith(self):
        query = Query().from_table(
            table='test_table'
        ).where(Q(
            field_name__startswith='some value'
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE (field_name LIKE %(A0)s)'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))
        self.assertEqual(query._where.args['A0'], 'some value%', 'Value is not correct')

    def test_where_not_startswith(self):
        query = Query().from_table(
            table='test_table'
        ).where(~Q(
            field_name__startswith='some value'
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE ((NOT(field_name LIKE %(A0)s)))'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))
        self.assertEqual(query._where.args['A0'], 'some value%', 'Value is not correct')

    def test_where_and(self):
        query = Query().from_table(
            table='test_table'
        ).where(Q(
            one='two'
        )).where(Q(
            three='four'
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE (one = %(A0)s AND three = %(A1)s)'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_and_combined(self):
        query = Query().from_table(
            table='test_table'
        ).where(Q(
            one='two'
        ) & Q(
            three='four'
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE (one = %(A0)s AND three = %(A1)s)'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_or(self):
        query = Query().from_table(
            table='test_table'
        ).where(Q(
            one='two'
        )).where(Q(
            three='four'
        ), OR)

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE ((one = %(A0)s) OR (three = %(A1)s))'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_combined_or(self):
        query = Query().from_table(
            table='test_table'
        ).where(Q(
            one='two'
        ) | Q(
            three='four'
        ))

        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table WHERE ((one = %(A0)s OR three = %(A1)s))'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_and_with_combined_or(self):
        query = Query().from_table(
            table='test_table'
        ).where(Q(
            one='two'
        )).where(Q(
            three='four'
        ) | Q(
            five='six'
        ))

        query_str = query.get_sql()
        expected_query = (
            'SELECT test_table.* '
            'FROM test_table '
            'WHERE (one = %(A0)s AND (three = %(A1)s OR five = %(A2)s))'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_and_with_not_combined_or(self):
        query = Query().from_table(
            table='test_table'
        ).where(Q(
            one='two'
        )).where(~Q(
            three='four'
        ) | Q(
            five='six'
        ))

        query_str = query.get_sql()
        expected_query = (
            'SELECT test_table.* '
            'FROM test_table '
            'WHERE (one = %(A0)s AND ((NOT(three = %(A1)s)) OR five = %(A2)s))'
        )
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_where_complex(self):
        query = Query().from_table(
            table='test_table'
        ).where(Q(
            one=1
        )).where(Q(
            two__gt=2
        )).where(~Q(
            three__gte=3
        )).where(~Q(
            four__lt=4
        ), OR).where(Q(
            five__lte=5
        ), OR).where(Q(
            six__contains='six'
        )).where(~Q(
            seven__startswith='seven'
        )).where(Q(
            eight=8
        ) & Q(
            nine=9
        ) | Q(
            ten=10
        ) | ~Q(
            eleven=11
        ))

        query_str = query.get_sql()
        expected_query = ''.join([
            'SELECT test_table.* FROM test_table WHERE ',
            '(((one = %(A0)s AND two > %(A1)s AND (NOT(three >= %(A2)s))) OR (NOT(four < %(A3)s)) ',
            'OR five <= %(A4)s) AND (six LIKE %(A5)s) AND (NOT(seven LIKE %(A6)s)) AND ',
            '((eight = %(A7)s AND nine = %(A8)s) OR ten = %(A9)s OR (NOT(eleven = %(A10)s))))'
        ])
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))



import abc

from six import string_types, with_metaclass


class FieldFactory(object):
    """
    Creates the correct field class based on the type of the passed field
    """

    def __new__(cls, field, *args, **kwargs):
        """
        Determines which type of field class to instantiate based on the field argument

        :param field: The field used in determining which type of Field object to return.
            This can be a string of the field name, a dict of {'alias': field},
            or a ``Field``
        :type table: str or dict or :class:`Field <querybuilder.fields.Field>`

        :return: The Field instance if a valid type was determined, otherwise None
        :rtype: :class:`Field <querybuilder.fields.Field>` or None
        """
        field_type = type(field)
        if field_type is dict:
            kwargs.update(alias=list(field.keys())[0])
            field = list(field.values())[0]
            field_type = type(field)

        if isinstance(field, string_types):
            return SimpleField(field, **kwargs)
        elif isinstance(field, Field):
            for key, value in kwargs.items():
                setattr(field, key, value)
            return field
        return None


class Field(with_metaclass(abc.ABCMeta, object)):
    """
    Abstract field class that all field types extend.

    Properties:

        name: str
            The name that identifies this table if there is no alias

        alias: str
            The optional alias used to identify this table

        auto_alias: str
            An alias that is set automatically by the Query if needed for inner query
            namespacing

        ignore: bool
            If set to True before the field is added to a table, this field will be
            ignored and not actually added to the table list. Typically used for fields
            that will create other fields like '*' or auto date fields.

        auto: bool
            This is a flag that is read when adding fields which could indicate some
            other fields need to be automatically created.
    """

    def __init__(self, field=None, table=None, alias=None, cast=None, distinct=None):
        """
        :param field: A string of a field name
        :type field: str

        :param table: A Table instance used to disambiguate the field. This is optional in
            simple queries
        :type table: :class:`Table <querybuilder.tables.Table>`

        :param alias: An alias to be used for this field
        :type alias: str

        :param cast: A data type name this field should be cast to. Ex: 'float'
        :type cast: str

        :param distinct: Indicates if a DISTINCT flag should be added during sql generation
        :type cast: bool
        """
        # TODO: implement distinct
        self.field = field
        self.name = None
        self.table = table
        self.alias = alias
        self.auto_alias = None
        self.ignore = False
        self.auto = False
        self.cast = cast
        self.distinct = distinct

    def get_sql(self):
        """
        Gets the SELECT sql part for a field
        Ex: field_name AS alias

        :return: the sql for this field used in the SELECT portion of the query
        :rtype: str
        """
        alias = self.get_alias()
        if alias:
            if self.cast:
                return 'CAST({0} AS {1}) AS {2}'.format(self.get_select_sql(), self.cast.upper(), alias)
            return '{0} AS {1}'.format(self.get_select_sql(), alias)

        if self.cast:
            return 'CAST({0} AS {1})'.format(self.get_identifier(), self.cast.upper())
        return self.get_identifier()

    def get_name(self):
        """
        Gets the name for the field and returns it. This identifies the field if there
        is not an alias set.

        :return: The name for the field
        :rtype: str
        """
        alias = self.get_alias()
        if alias:
            return alias
        return self.name

    def get_alias(self):
        """
        Gets the alias for the field or the auto_alias if one is set.
        If there isn't any kind of alias, None is returned.

        :return: The field alias, auto_alias, or None
        :rtype: str or None
        """
        alias = None
        if self.alias:
            alias = self.alias
        elif self.auto_alias:
            alias = self.auto_alias

        if self.table and self.table.prefix_fields:
            field_prefix = self.table.get_field_prefix()
            if alias:
                alias = '{0}__{1}'.format(field_prefix, alias)
            else:
                alias = '{0}__{1}'.format(field_prefix, self.name)

        return alias

    def get_identifier(self):
        """
        Gets the name for the field of how it should be referenced within a query.
        It will be prefixed with the table name or table alias

        :return: the name to reference the field within a query
        :rtype: str
        """
        alias = self.get_alias()
        if alias:
            return alias
        return self.get_select_sql()

    def get_select_sql(self):
        """
        Gets the SELECT field portion for the field without the alias. If the field
        has a table, it will be included here like table.field

        :return: Gets the SELECT field portion for the field without the alias
        :rtype: str
        """
        if self.table:
            return '{0}.{1}'.format(self.table.get_identifier(), self.name)
        return '{0}'.format(self.name)

    def before_add(self):
        """
        Template method to be implemented by subclasses. This is called
        before the field is actually added to a table
        """
        pass

    def set_table(self, table):
        """
        Setter for the table. This is meant to be extended by any subclass that might need
        to do additional processing with the table it belongs to. Ex: aggregate functions
        which reference multiple fields can set their inner fields' table.
        """
        self.table = table


class SimpleField(Field):
    """
    A field that is created with just the string name of the field
    """

    def __init__(self, field=None, table=None, alias=None, cast=None, distinct=None):
        """
        Sets the name of the field to the passed in field value

        :param field: A string of a field name
        :type field: str

        :param table: A Table instance used to disambiguate the field. This is optional in
            simple queries
        :type table: :class:`Table <querybuilder.tables.Table>`

        :param alias: An alias to be used for this field
        :type alias: str

        :param cast: A data type name this field should be cast to. Ex: 'float'
        :type cast: str

        :param distinct: Indicates if a DISTINCT flag should be added during sql generation
        :type cast: bool
        """
        super(SimpleField, self).__init__(field, table, alias, cast, distinct)
        self.name = field


class JsonField(SimpleField):
    """
    Used to access postgres native json fields
    """
    def __init__(self, field=None, table=None, alias=None, cast=None, distinct=None, key=None):
        super(JsonField, self).__init__(field, table, alias, cast, distinct)
        self.key = key

    def get_select_sql(self):
        if self.table:
            return '{0}.{1}->\'{2}\''.format(self.table.get_identifier(), self.name, self.key)
        return '{0}->\'{1}\''.format(self.name, self.key)

    def get_where_key(self):
        if self.table:
            return '{0}.{1}->>\'{2}\''.format(self.table.get_identifier(), self.name, self.key)
        return '{0}->>\'{1}\''.format(self.name, self.key)


class MultiField(Field):
    """
    A field that contains one or more nested fields
    """

    def __init__(self, field=None, table=None, alias=None, cast=None, distinct=None):
        # TODO: implement handling of more than one nested field
        super(MultiField, self).__init__(field, table, alias, cast, distinct)
        self.field = FieldFactory(field)
        if self.table:
            self.field.set_table(self.table)

    def set_table(self, table):
        """
        Setter for the table of this field. Also sets the inner field's table.
        """
        super(MultiField, self).set_table(table)
        if self.field and self.field.table is None:
            self.field.set_table(self.table)

    def get_field_identifier(self):
        """
        Gets the identifier of the nested field

        :return: the identifier of the nested field
        :rtype: str
        """
        return self.field.get_identifier()


class ExpressionField(Field):
    pass


class AggregateField(MultiField):
    """
    The base class for aggregate functions and window functions.

    Properties:

        function_name: str
            The aggregate function name. This is used to automatically generate the sql
            for simple aggregate functions.
    """
    function_name = None

    def __init__(self, field=None, table=None, alias=None, cast=None, distinct=None, over=None):
        """
        Sets the field to a field instance because aggregate functions are treated as fields
        that perform an operation on a db column

        :param field: A string of a field name
        :type field: str

        :param table: A Table instance used to disambiguate the field. This is optional in
            simple queries
        :type table: :class:`Table <querybuilder.tables.Table>`

        :param alias: An alias to be used for this field
        :type alias: str

        :param cast: A data type name this field should be cast to. Ex: 'float'
        :type cast: str

        :param distinct: Indicates if a DISTINCT flag should be added during sql generation
        :type cast: bool

        :param over: The QueryWindow to perform the aggregate function over
        :type over: :class:`QueryWindow <querybuilder.query.QueryWindow>`
        """
        super(AggregateField, self).__init__(field, table, alias, cast, distinct)
        self.name = self.function_name
        self.over = over

        field_name = None
        if self.field and isinstance(self.field.field, string_types):
            field_name = self.field.field
            if field_name == '*':
                field_name = 'all'

        if field_name:
            self.auto_alias = '{0}_{1}'.format(field_name, self.name.lower())
        else:
            self.auto_alias = self.name.lower()

    def get_select_sql(self):
        """
        Gets the SELECT field portion for the field without the alias. If the field
        has a table, it will be included here like AggregateFunction(table.field)

        :return: Gets the SELECT field portion for the field without the alias
        :rtype: str
        """
        return '{0}({1}){2}'.format(
            self.name.upper(),
            self.get_field_identifier(),
            self.get_over(),
        )

    def get_field_identifier(self):
        """
        Gets the identifier of the field used in the aggregate function

        :return: the identifier of the field used in the aggregate function
        :rtype: str
        """
        return self.field.get_identifier()

    def get_over(self):
        """
        Gets the over clause to be used in the window function sql

        :returns: the over clause to be used in the window function sql
        :rtype: str
        """
        if self.over:
            return ' {0}'.format(self.over.get_sql())
        return ''


class CountField(AggregateField):
    """
    Count aggregation
    """
    function_name = 'Count'


class AvgField(AggregateField):
    """
    Average aggregation
    """
    function_name = 'Avg'


class MaxField(AggregateField):
    """
    Maximum aggregation
    """
    function_name = 'Max'


class MinField(AggregateField):
    """
    Minimum aggregation
    """
    function_name = 'Min'


class StdDevField(AggregateField):
    """
    Standard deviation aggregation
    """
    function_name = 'StdDev'


class NumStdDevField(AggregateField):
    """
    Number of standard deviations from the average aggregation
    """
    function_name = 'num_stddev'

    def get_select_sql(self):
        """
        To calculate the number of standard deviations calculate the difference
        of the field and the average and divide the difference by the standard
        deviation
        """
        return '(CASE WHEN (STDDEV({0}){1}) <> 0 THEN (({0} - (AVG({0}){1})) / (STDDEV({0}){1})) ELSE 0 END)'.format(
            self.get_field_identifier(),
            self.get_over(),
        )


class SumField(AggregateField):
    """
    Summation aggregation
    """
    function_name = 'Sum'


class VarianceField(AggregateField):
    """
    Variance window function
    """
    function_name = 'Variance'


class RowNumberField(AggregateField):
    """
    Row number window function
    """
    function_name = 'row_number'

    def get_field_identifier(self):
        return ''


class RankField(AggregateField):
    """
    Rank window function
    """
    function_name = 'rank'

    def get_field_identifier(self):
        return ''


class DenseRankField(AggregateField):
    """
    Dense rank window function
    """
    function_name = 'dense_rank'

    def get_field_identifier(self):
        return ''


class PercentRankField(AggregateField):
    """
    Percent rank window function
    """
    function_name = 'percent_rank'

    def get_field_identifier(self):
        return ''


class CumeDistField(AggregateField):
    """
    Cume dist window function
    """
    function_name = 'cume_dist'

    def get_field_identifier(self):
        return ''


class NTileField(AggregateField):
    """
    NTile window function
    """
    function_name = 'ntile'

    def __init__(self, field=None, table=None, alias=None, cast=None, distinct=None, over=None, num_buckets=1):
        """
        Sets the num_buckets for ntile

        :param field: A string of a field name
        :type field: str

        :param table: A Table instance used to disambiguate the field. This is optional in
            simple queries
        :type table: :class:`Table <querybuilder.tables.Table>`

        :param alias: An alias to be used for this field
        :type alias: str

        :param cast: A data type name this field should be cast to. Ex: 'float'
        :type cast: str

        :param distinct: Indicates if a DISTINCT flag should be added during sql generation
        :type cast: bool

        :param over: The QueryWindow to perform the aggregate function over
        :type over: :class:`QueryWindow <querybuilder.query.QueryWindow>`

        :param num_buckets: Number of buckets to use for ntile
        :type num_buckets: int
        """
        super(NTileField, self).__init__(field, table, alias, cast, distinct, over)
        self.num_buckets = num_buckets

    def get_field_identifier(self):
        """
        Returns the number of buckets

        :return: the number of buckets used for the ntile function
        :rtype: int
        """
        return self.num_buckets


class LeadLagField(AggregateField):
    """
    Base class for lag and lead window functions
    """

    def __init__(self, field=None, table=None, alias=None, cast=None, distinct=None, over=None, offset=1, default=None):
        """
        Sets the offset and default value for the lag/lead calculation

        :param field: A string of a field name
        :type field: str

        :param table: A Table instance used to disambiguate the field. This is optional in
            simple queries
        :type table: :class:`Table <querybuilder.tables.Table>`

        :param alias: An alias to be used for this field
        :type alias: str

        :param cast: A data type name this field should be cast to. Ex: 'float'
        :type cast: str

        :param distinct: Indicates if a DISTINCT flag should be added during sql generation
        :type cast: bool

        :param over: The QueryWindow to perform the aggregate function over
        :type over: :class:`QueryWindow <querybuilder.query.QueryWindow>`

        :param offset: The offset number of rows which to calculate the lag/lead
        :type offset: int

        :param default: The default value to use if the offset doesn't find a field
        :type default: number or str or object
        """
        super(LeadLagField, self).__init__(field, table, alias, cast, distinct, over)
        self.offset = offset
        self.default = default

    def get_field_identifier(self):
        """
        Return the lag/lead function with the offset and default value
        """
        if self.default is None:
            return '{0}, {1}'.format(self.field.get_select_sql(), self.offset)
        return "{0}, {1}, '{2}'".format(self.field.get_select_sql(), self.offset, self.default)


class LagField(LeadLagField):
    """
    Lag window function
    """
    function_name = 'lag'


class LeadField(LeadLagField):
    """
    Lead window function
    """
    function_name = 'lead'


class LeadLagDifferenceField(LeadLagField):
    """
    Base class for lag difference and lead difference window functions
    """
    def get_select_sql(self):
        """
        Calculate the difference between this record's value and the lag/lead record's value
        """
        return '(({0}) - ({1}({2}){3}))'.format(
            self.field.get_select_sql(),
            self.name.upper(),
            self.get_field_identifier(),
            self.get_over(),
        )


class LagDifferenceField(LeadLagDifferenceField):
    """
    Lag difference window function
    """
    function_name = 'lag'


class LeadDifferenceField(LeadLagDifferenceField):
    """
    Lead difference window function
    """
    function_name = 'lead'


class FirstValueField(AggregateField):
    """
    First value window function
    """
    function_name = 'first_value'


class LastValueField(AggregateField):
    """
    Last value window function
    """
    function_name = 'last_value'


class NthValueField(AggregateField):
    """
    Nth value window function
    """
    function_name = 'nth_value'

    def __init__(self, field=None, table=None, alias=None, cast=None, distinct=None, over=None, n=1):
        """
        Sets the Nth value

        :param field: A string of a field name
        :type field: str

        :param table: A Table instance used to disambiguate the field. This is optional in
            simple queries
        :type table: :class:`Table <querybuilder.tables.Table>`

        :param alias: An alias to be used for this field
        :type alias: str

        :param cast: A data type name this field should be cast to. Ex: 'float'
        :type cast: str

        :param distinct: Indicates if a DISTINCT flag should be added during sql generation
        :type cast: bool

        :param over: The QueryWindow to perform the aggregate function over
        :type over: :class:`QueryWindow <querybuilder.query.QueryWindow>`

        :param n: the n value to use for the Nth value function
        :type n: int
        """
        super(NthValueField, self).__init__(field, table, alias, cast, distinct, over)
        self.n = n

    def get_field_identifier(self):
        """
        Returns the field's sql and the n value

        :return: the field's sql and the n value
        :rtype: str
        """
        return '{0}, {1}'.format(self.field.get_select_sql(), self.n)


class DatePartField(MultiField):
    """
    Base class for all date part fields

    Properties:

        group_name: str
            The name of the date part
    """
    group_name = None

    def __init__(self, field=None, table=None, alias=None, cast=None, distinct=None, auto=False, desc=False,
                 include_datetime=False):
        """
        :param field: A string of a field name
        :type field: str

        :param table: A Table instance used to disambiguate the field. This is optional in
            simple queries
        :type table: :class:`Table <querybuilder.tables.Table>`

        :param alias: An alias to be used for this field
        :type alias: str

        :param cast: A data type name this field should be cast to. Ex: 'float'
        :type cast: str

        :param distinct: Indicates if a DISTINCT flag should be added during sql generation
        :type cast: bool

        :param auto: Set to True for this date field to auto generate the necessary other date parts and
            apply the appropriate groupings. Defaults to False
        :type auto: bool

        :param desc: If this is set to True, the results will be sorted by the datetime field in
            descending order. Defaults to False
        :type desc: bool

        :param include_datetime: If set to True, datetime objects will be returned in addition to the
            unix timestamp. Defaults to False.
        :type include_datetime: bool
        """
        super(DatePartField, self).__init__(field, table, alias, cast, distinct)
        self.name = self.group_name
        self.auto = auto
        self.desc = desc
        self.include_datetime = include_datetime

        if self.cast is None:
            self.cast = 'INT'

        self.auto_alias = '{0}__{1}'.format(self.field.get_name(), self.name)

    def get_select_sql(self):
        """
        Extract the date part from the date field

        :return: The EXTRACT sql portion for this field
        :rtype: str
        """
        return 'EXTRACT({0} FROM {1})'.format(self.name, self.field.get_select_sql())

    def before_add(self):
        """
        Generate auto fields for this date part if auto is True.
        """
        if self.auto:
            self.ignore = True
            self.generate_auto_fields()

    def generate_auto_fields(self):
        """
        Generates any auto fields needed to properly group this date part.
        Ex: a Day field will create Year, Month, Day fields and group by Year, Month, Day
        """
        # ignore the original date field
        self.ignore = True
        datetime_str = None

        # create an alias for the unix timestamp extraction
        epoch_alias = '{0}__{1}'.format(self.field.get_name(), 'epoch')

        if self.name == 'all':
            datetime_str = self.field
            self.add_to_table(AllEpoch(datetime_str, table=self.table, cast=self.cast), epoch_alias)
            # do not add the date order by for "all" grouping because we want to order by rank
            return
        elif self.name == 'none':
            datetime_str = self.field
            self.add_to_table(Epoch(datetime_str, table=self.table, cast=self.cast), epoch_alias, add_group=True)
        else:
            group_names = default_group_names
            if self.name == 'week':
                group_names = week_group_names

            group_name_index = group_names.index(self.name) + 1
            for group_name in group_names[0:group_name_index]:
                field_alias = '{0}__{1}'.format(self.field.get_name(), group_name)
                auto_field = group_map[group_name](self.field, table=self.table, cast=self.cast)
                self.add_to_table(auto_field, field_alias, add_group=True)

                # check if this is the last date grouping
                if group_name == self.name:
                    datetime_str = self.field
                    self.add_to_table(
                        GroupEpoch(
                            datetime_str,
                            date_group_name=group_name,
                            table=self.table,
                            cast=self.cast,
                        ),
                        epoch_alias,
                        add_group=True
                    )

        if self.desc:
            self.table.owner.order_by('-{0}'.format(epoch_alias))
        else:
            self.table.owner.order_by(epoch_alias)

    def add_to_table(self, field, alias, add_group=False):
        """
        Adds this field to the field's table and optionally group by it

        :param field: The field to add to the table
        :type field: str or :class:`Field <querybuilder.fields.Field>`

        :param alias: The alias for the field
        :type alias: str

        :param add_group: Whether or not the table should group by this field
        :type: bool
        """
        self.table.add_field({
            alias: field
        })
        if add_group:
            self.table.owner.group_by(alias)


class AllTime(DatePartField):
    """
    Date group will be over the whole time range
    """
    group_name = 'all'

    def __init__(self, field=None, table=None, alias=None, cast=None, distinct=None, auto=False, desc=False,
                 include_datetime=False):
        super(AllTime, self).__init__(field, table, alias, cast, distinct, auto, desc, include_datetime)
        self.auto = True


class NoneTime(DatePartField):
    """
    No date grouping is used
    """
    group_name = 'none'

    def __init__(self, field=None, table=None, alias=None, cast=None, distinct=None, auto=False, desc=False,
                 include_datetime=False):
        super(NoneTime, self).__init__(field, table, alias, cast, distinct, auto, desc, include_datetime)
        self.auto = True


class Year(DatePartField):
    """
    Extract the year from the datetime
    """
    group_name = 'year'


class Month(DatePartField):
    """
    Extract the month from the datetime
    """
    group_name = 'month'


class Day(DatePartField):
    """
    Extract the day from the datetime
    """
    group_name = 'day'


class Hour(DatePartField):
    """
    Extract the hour from the datetime
    """
    group_name = 'hour'


class Minute(DatePartField):
    """
    Extract the minute from the datetime
    """
    group_name = 'minute'


class Second(DatePartField):
    """
    Extract the second from the datetime
    """
    group_name = 'second'


class Week(DatePartField):
    """
    Extract the week from the datetime
    """
    group_name = 'week'


class Epoch(DatePartField):
    """
    Extract the epoch from the datetime
    """
    group_name = 'epoch'

    def __init__(self, field=None, table=None, alias=None, cast=None, distinct=None, auto=False, desc=False,
                 include_datetime=False, date_group_name=None):
        self.date_group_name = date_group_name
        super(Epoch, self).__init__(field, table, alias, cast, distinct, auto, desc, include_datetime)


class GroupEpoch(Epoch):
    """
    Epoch used for a date grouping
    """

    def get_select_sql(self):
        return 'EXTRACT({0} FROM date_trunc(\'{1}\', {2}))'.format(
            self.name,
            self.date_group_name,
            self.field.get_identifier()
        )


class AllEpoch(Epoch):
    """
    Epoch used for all grouping
    """
    def get_select_sql(self):
        return 0


group_map = {
    'year': Year,
    'month': Month,
    'day': Day,
    'hour': Hour,
    'minute': Minute,
    'second': Second,
    'week': Week,
    'all': AllTime,
    'none': NoneTime,
}

all_group_names = (
    'year',
    'month',
    'day',
    'hour',
    'minute',
    'second',
    'week',
    'all',
    'none',
)

allowed_group_names = (
    'year',
    'month',
    'day',
    'hour',
    'minute',
    'second',
    'week',
)

default_group_names = (
    'year',
    'month',
    'day',
    'hour',
    'minute',
    'second',
)

week_group_names = (
    'year',
    'week',
)



from querybuilder.query import Query
from querybuilder.tests.query_tests import QueryTestCase, get_comparison_str


class LimitTest(QueryTestCase):
    def test_limit(self):
        query = Query().from_table(
            table='test_table'
        ).limit(10)
        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table LIMIT 10'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_offset(self):
        query = Query().from_table(
            table='test_table'
        ).limit(
            offset=10
        )
        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table OFFSET 10'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))

    def test_limit_with_offset(self):
        query = Query().from_table(
            table='test_table'
        ).limit(
            limit=5,
            offset=20
        )
        query_str = query.get_sql()
        expected_query = 'SELECT test_table.* FROM test_table LIMIT 5 OFFSET 20'
        self.assertEqual(query_str, expected_query, get_comparison_str(query_str, expected_query))
